/* MPROM Firmware Code version 1.0 (Debug monitor)                   */
/* programmed by M.yanagisawa(MoroboshiLUM)  elfin@mth.biglobe.ne.jp */
/* Copyright (c) 2000 by M.yanagisawa. All rigths reserved.          */

#ifdef __BORLANDC__

#include <stdio.h>
#include <dos.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <alloc.h>

#else

#include <stdrom.h>

#endif

#include "generic.h"
#include "drawlib.h"
#include "rslib.h"
#include "mprom.h"
#include "i2clib.h"
#include "main.h"
#include "atactrl.h"

/* command list */
#define _NOP 255
enum {_INPORTB,_INPORTW,_OUTPORTB,_OUTPORTW,_DUMP,_FILL,_EXEC,_CLEAR,_LOAD,
      _EJECT,_RESET,_READTOC,_RSENSE,_READDATA,_DIR,_PLAY,_STOP,_SENDDAC,_CHDIR,
      _LOADDATA,_SENDDSP,_SETFEATURE,_PLAYDATA,_DISP,_DISPG,_DISPD,_ID3,_QUIT };

const char *cmdstr[]={"IB","IW","OB","OW","D","F","G","C","L",
                      "EJECT","RESET","TOC","SENSE","RD","DIR","PLAY","STOP","SENDDAC",
                      "CD","LD","SP","SETF","PLAYM","DISP","DISPG","DISPD","ID3","Q",NULL};

static WORD _DEFSEG,_DEFOFS;

char *getparam(char *s,WORD *n)
{
   char *t;
   int siz;

   t=s;
   siz=strspn(s," "); /* skip space */
   if (strlen(s)==siz) return(NULL); /* err!  */
   t=t+siz;
   *n=0; /* clear */
   while (isxdigit(*t)) {
     if (isalpha(*t)) *n=*n*0x10+(*t-'A'+0xA);
     else	*n=*n*0x10+(*t-'0');
     t++;
   }
   if (*t && (*t!=' ')) return(NULL); /* else space or 0x00 */
   return(t);
}

char *getparam10(char *s,WORD *n)
{
   char *t;
   int siz;

   t=s;
   siz=strspn(s," "); /* skip space */
   if (strlen(s)==siz) return(NULL); /* err!  */
   t=t+siz;
   *n=0; /* clear */
   while (isdigit(*t)) {
     *n=*n*10+(*t-'0');
     t++;
   }
   if (*t && (*t!=' ')) return(NULL); /* else space or 0x00 */
   return(t);
}

int getcmd(char **s,char *buf)
{
   char *t,*u;
   int siz,i=0;
  
   t=buf;
   siz=strspn(buf," "); /* skip space */
   if (strlen(buf)==siz) return(_NOP); /* err!  */
   t=t+siz;
   u=strchr(t,' ');
   if (u!=NULL) {
     *u=0x00; /* null */
     u++;
   }
   *s=u;  /* next str */
   while (cmdstr[i]!=NULL) {
     if (!strcmp(t,cmdstr[i])) return(i);
     i++;
   }
   return(-1);
}

void dispparamerr(char *s)
{
   if (s!=NULL)
     sprintf(dispbuf,"param err: %s\r\n",s);
   else
     strcpy(dispbuf,"param err\r\n");
   putmsg(dispbuf);
}

void _EXEC_DISP(int cmd,char *s)
{
   WORD d1,d2;
   char *s1,*s2;
 
   s1=getparam10(s,&d1);
   if (s1!=NULL) {
      s2=getparam10(s1,&d2);
      if (s2==NULL) {
        dispparamerr(s);
        return;
      }
   } else {
     dispparamerr(s);
     return;
   }
   if (d1>80) return;
   if (d2>14) return;
   if (*s2) s2++;
#ifdef __BORLANDC__

	_asm {
		mov ax,12h
		int 10h
	}
	_disable();
	outpw(0x3ce,0x0005); /* writemode 0    */
	outpw(0x3ce,0x0001); /* Latch _disable */
	outpw(0x3ce,0xff08); /* all bit enable */
	outpw(0x3ce,0x0003); /* set mode/rotate=0 */
    outpw(0x3c4,0x0f02); /* sequencer */
    _enable();
	Grph_Init();
	messagedlg("zikkou simasuka?",_MSGDLG_1BTN);
	getch();
	_asm {
	  mov	ax,3h
	  int	10h
	}
#else
	Grph_Init();
    putmessage(d1,d2,s2);
#endif
}

void _EXEC_INPORT(int cmd,char *s)
{
   WORD d,n;

   if (getparam(s,&d)) {
      if (cmd==_INPORTW) n=inpw(d);
      else n=inp(d);
      sprintf(dispbuf,"port $%X = $%X\r\n",d,n);
      putmsg(dispbuf);
   }
   else 
     dispparamerr(s);
}

void _EXEC_PLAY(int cmd,char *s)
{
   WORD d;
   DWORD len,sta;

   if (getparam10(s,&d)) {
     if ((d==0) || (d>TrackNum)) return;
	 sta=TrackLba[d];
     len=TrackLba[0]-sta;
     PlayCD(sta,len);
   }
}

void _EXEC_OUTPORT(int cmd,char *s)
{
   WORD d1,d2;
   char *s1;

   s1=getparam(s,&d1);
   if (s1!=NULL) {
      if (getparam(s1,&d2)) {
        if (cmd==_OUTPORTW) outpw(d1,d2);
        else outp(d1,d2);
        sprintf(dispbuf,"write $%X to port $%X\r\n",d2,d1);
        putmsg(dispbuf);
        return;
      }
   }
   dispparamerr(s);
}

void dump_core(WORD seg,WORD ofs,int cnt,BOOL isglobal)
{
   int i,j;
   BYTE far *p;

   for (i=0;i<cnt;i++) {
     sprintf(dispbuf,"%.4X:%.4X ",seg,ofs);
     putmsg(dispbuf);
     p=(BYTE far *)MK_FP(seg,ofs);
     for (j=0;j<16;j++) {
       sprintf(dispbuf,"%.2X ",*p++);
       putmsg(dispbuf);
       if (ofs==0xffff) {
         ofs=0; seg=seg+0x1000;
       } 
         else ofs++;
     }
     putmsg("\r\n");
   }
   if (isglobal) {
     _DEFSEG=seg; _DEFOFS=ofs;
   }
}

void _EXEC_DUMP(int cmd,char *s)
{
   WORD d1,d2,d3,d4,seg,ofs,toseg,toofs;
   char *s1,*s2,*s3;
   long cnt;

   cnt=8;
   s1=s2=s3=NULL;
   s1=getparam(s,&d1);
   if (s1!=NULL) {
      s2=getparam(s1,&d2);
   }
   if ((s1==NULL) && (s2==NULL)) {
     seg=_DEFSEG; ofs=_DEFOFS;
   } else if (s2==NULL) {
     /* offset only */
     seg=_DEFSEG;
     ofs=d1;
   } else {
     seg=d1; ofs=d2;
   }
   if (s2!=NULL) s3=getparam(s2,&d3);
   if (s3!=NULL) { 
     getparam(s3,&d4);
     toseg=d3; toofs=d4 & 0xfff0;
     cnt=(DWORD)toseg*16+(DWORD)toofs-(DWORD)seg*16+(DWORD)ofs;
     if (cnt<0) {
       dispparamerr(s);
       return;
     }
     cnt=cnt >> 4;
   }
   ofs&=0xfff0; // 
   dump_core(seg,ofs,cnt,TRUE);
}

void _EXEC_FILL(int cmd,char *s)
{
   WORD d[4],seg,ofs,cnt;
   char *ss[4];
   BYTE far *p;
   BYTE dat;
   int i;

   for (i=0;i<4;i++) ss[i]=NULL;
   ss[0]=getparam(s,&d[0]);
   for (i=1;i<4;i++)
     if (ss[i-1]!=NULL) ss[i]=getparam(ss[i-1],&d[i]);
   for (i=0;i<4;i++)
     if (ss[i]==NULL) {
       dispparamerr(s);
       return;
     }
   seg=d[0];
   ofs=d[1] & 0xfff0;
   cnt=d[2]; dat=d[3];
   p=(BYTE far *)MK_FP(seg,ofs);
   for (i=0;i<cnt;i++) *p++=dat;
   sprintf(dispbuf,"%.2X : %X bytes filled from %.4X:%.4X\r\n",dat,cnt,seg,ofs);
   putmsg(dispbuf);
}

void _EXEC_PROG(int cmd,char *s)
{
   WORD d1,d2;
   char *s1;
   void (far *p)(void);

   s1=getparam(s,&d1);
   if (s1!=NULL) {
      if (getparam(s1,&d2)) {
        sprintf(dispbuf,"Execute program at %.4X:%.4X ...\r\n",d1,d2);
        putmsg(dispbuf);
		p=(void (far *)())MK_FP(d1,d2);
		p();
      }
   }
   dispparamerr(s);
}

BYTE convbyte(char c1,char c2)
{
  BYTE n;

  if (isdigit(c1)) n=(c1-'0') << 4;
  else             n=(c1-'A'+0xA) << 4;
  if (isdigit(c2)) n|=c2-'0';
  else             n|=(c2-'A'+0xA);
  return(n);
}

/* get base ZOgx[X擾 */
WORD getsegbase(char *s)
{
  if (strlen(s)>0)
   if (*s==':')
     return(((WORD)convbyte(s[9],s[10]) << 8) | (WORD)convbyte(s[11],s[12]));
  return(0);
}

/* write data-field(sum check) */
int writedata(WORD segadr,char *s)
{
  BYTE c,d,sum;
  int err,i,max,j;
#ifndef __BORLANDC__
  WORD ofs;
  BYTE far *p;
#endif

  err=0; sum=0;
  if (!strlen(s)) return(-1);
  if (s[0]==':') {
    /* delete illegal data */
    for (i=1;i<strlen(s);i++)
     if (!isxdigit(s[i])) err=-1;
    if (strlen(s)<12) err=-1;
    max=convbyte(s[1],s[2]);
    if (strlen(s)!=(max*2)+11) err=-1;
    if (!err) {
#ifndef __BORLANDC__
      ofs=((WORD)convbyte(s[3],s[4]) << 8) |  (WORD)convbyte(s[5],s[6]);
  	  p=(BYTE far *)MK_FP(segadr,ofs);
#endif
      /* calc sum 5bytes */
      for (j=1,i=0;i<4;j+=2,i++)
        sum+=convbyte(s[j],s[j+1]);
      for (i=0;i<max;i++,j+=2) {
        c=convbyte(s[j],s[j+1]);  sum+=c;
#ifndef __BORLANDC__
  	    *p++=c;
#endif
      }
      c=convbyte(s[j],s[j+1]);
      d=~sum+1;
      if (d!=c) err=-1;
    }
  }
  return(err);
}


void _EXEC_LOAD(int cmd,char *s)
{
  WORD segadr;
  char buf[256];
  int cnt,err;
#ifdef __BORLANDC__
  FILE *fp;
#endif

  segadr=0; err=cnt=0;
  putmsg("Waiting uploading...\r\n");
#ifdef __BORLANDC__
  if ((fp=fopen("dummy.hex","r"))==NULL) return;
	while (!feof(fp)) {
	  fgets(buf,sizeof(buf)-1,fp);
	  if (strlen(buf)) buf[strlen(buf)-1]=0;
#else
    while (1) {
      getmsg(0,buf);
#endif
	  if (strchr(buf,_ESC)!=NULL) {  
	    err=-1;
	    break; /*ESC*/
	  }
      cnt++;
      if (cnt & 1) putmsg(".");
	  if (strstr(buf,":02000002")==buf) segadr=getsegbase(buf);
	  else if (strstr(buf,":00000001FF")==buf) break;
	  else {
	    if (writedata(segadr,buf)) {
	      putmsg("load hex file err\r\n");
	      err=1;
	      break;
	    }
      }
   }
#ifdef __BORLANDC__
  fclose(fp);
#endif
  if (!err) putmsg("\r\ndone.\r\n");
  else if (err<0) putmsg("\r\nbreak!\r\n");
}


void _EXEC_CLEAR(int cmd,char *s)
{
#ifndef __BORLANDC__
   int i;
   WORD seg;
   void far *p;

   seg=0x1000;
   for (i=1;i<=7;i++) {
     p=(void far *)MK_FP(seg,0);
     far_memset(p,0xff,0x8000);
     p=(void far *)MK_FP(seg+0x800,0);
     far_memset(p,0xff,0x8000);
     seg+=0x1000;
   }
#endif
}

void dispstat(void)
{
  if ((ERROR_STATUS & _ERR)==0) putmsg("ok\r\n");
  else putmsg("ng\r\n");
}


void _EXEC_EJECT(int cmd,char *s)
{
   ATAPIPAC pac;
   
   clratapipac(&pac);
   pac.cmd=0x1b;
   pac.param[3]=0x02;
#ifdef __BORLANDC__
   SendATAPICommand(_NODATA,&pac,12,NULL,0);
#else
   SendATAPICommand_INTR(&pac,12,NULL,0);
#endif
   dispstat();
}

void _EXEC_STOP(int cmd,char *s)
{
   PlayStop();
}

void _EXEC_READTOC(int cmd,char *s)
{
   BYTE far *buf;
   ATAPIPAC pac;
   int i,tr;
   
   TrackNum=0;
   buf=LocalBuf;
   if (!TestUnitReady()) {
     putmsg("device not ready\r\n");
     return;
   } else 
   if (!TestUnitReady()) {
     putmsg("device not ready\r\n");
     return;
   }
   clratapipac(&pac);
   pac.cmd=0x43;
   pac.param[5]=0x01;
   pac.param[6]=0x10;
   pac.param[7]=0x00;
#ifdef __BORLANDC__
   if (!SendATAPICommand(_DATAREAD,&pac,12,buf,0x1000)) {
     RequestSense(NULL);
     return;
   }
#else
   if (!SendATAPICommand_INTR(&pac,12,buf,0x1000)) {
     RequestSense(NULL);
     return;
   }
#endif
   TrackNum=buf[3];
   for (i=tr=0;i<=TrackNum;i++) {
     if (tr==TrackNum) tr=0;
     else			   tr=i+1;
     TrackLba[tr]=(DWORD)buf[4+i*8+7]+((DWORD)buf[4+i*8+6]<<8)
                +((DWORD)buf[4+i*8+5]<<16)+((DWORD)buf[4+i*8+4]<<24);
     if (tr) {
       if (buf[4+i*8+1] & 0x04) sprintf(dispbuf,"Data Track %d pos = %lx.\r\n",tr,TrackLba[tr]);
       else sprintf(dispbuf,"Audio Track %d pos = %lx.\r\n",tr,TrackLba[tr]);
       putmsg(dispbuf);
     }
   }
}
void _EXEC_REQUESTSENSE(int cmd,char *s)
{
   WORD seg,ofs;
   WORD buf[256];

   memset(buf,0,sizeof(buf));
   RequestSense((BYTE far *)buf);
   seg=FP_SEG((void far*)buf);
   ofs=FP_OFF((void far*)buf);
   dump_core(seg,ofs,16,FALSE);
   dispstat();
}

void _EXEC_RESET(int cmd,char *s)
{
  if (ATAReset()) putmsg("ok\r\n");
  else putmsg("ng\r\n");
}

void _EXEC_READDATA(int cmd,char *s)
{
   WORD seg,ofs;
   WORD d1;
   WORD buf[1024];
   DWORD lba;
   char *s1;
   ATAPIPAC pac;
#ifdef __BORLANDC__
   int siz;
   char *t;
   FILE *fp;
#endif
 
   memset(buf,0,sizeof(buf));
   s1=getparam(s,&d1);
   if (s1==NULL) return;
   lba=d1;
#ifdef __BORLANDC__
   t=s1;
   siz=strspn(s1," "); /* skip space */
   if (strlen(s1)==siz) return; /* err!  */
   t=t+siz;
   if ((fp=fopen(t,"wb"))==NULL) return;
#endif
   sprintf(dispbuf,"read at %x\r\n",lba);
   putmsg(dispbuf);
   clratapipac(&pac);
   pac.cmd=0xa8;
   pac.param[1]=(lba >> 24) & 0xff;
   pac.param[2]=(lba >> 16) & 0xff;
   pac.param[3]=(lba >> 8) & 0xff;
   pac.param[4]=lba & 0xff;
   pac.param[7]=0;
   pac.param[8]=1;
#ifdef __BORLANDC__
   SendATAPICommand(_DATAREAD,&pac,12,(BYTE far *)buf,sizeof(buf));
#else
   SendATAPICommand_INTR(&pac,12,(BYTE far *)buf,sizeof(buf));
#endif
#ifdef __BORLANDC__
   fwrite(buf,sizeof(buf),1,fp);
#endif
   seg=FP_SEG((void far*)buf);
   ofs=FP_OFF((void far*)buf);
   dump_core(seg,ofs,32,FALSE);
#ifdef __BORLANDC__
   fclose(fp);
#endif
   dispstat();
}

void DispType(int n,char *s,LPFILETYPE p)
{
   char typebuf[16];
   char filebuf[66];

   if (p->flag & 2) strcpy(typebuf,"<DIR>    ");
   else strcpy(typebuf,"         ");
   fstrcpy(filebuf,p->filename);
   if (filebuf[0]==1) strcpy(filebuf,"<ParentDir>");
   else if (filebuf[0]==0) strcpy(filebuf,"<CurrentDir>");
   sprintf(s,"%.2d. %-16s %s %lx",n,filebuf,typebuf,p->lba);
}

/* Directory Analizing (extract directory sector data to file info structures) */
BOOL DispDirectory(DWORD lba,DWORD length,BOOL isJoliet)
{
   LPFILETYPE ptr;
   int i;
   
   if (!LoadDirectory(lba,length,isJoliet)) return(FALSE);
   if (!DirFiles) return(FALSE);

   ptr=(LPFILETYPE)DirTbl;
   for (i=1;i<=DirFiles;i++) {
     DispType(i,dispbuf,ptr);
     putmsg(dispbuf);
     putmsg("\r\n");
     ptr=(LPFILETYPE)((BYTE far *)ptr+(DWORD)sizeof(FILETYPE));
   }
   putmsg("\r\n");
   return(TRUE);
}

void _EXEC_CHDIR(int cmd,char *s)
{
   WORD d1;
   LPFILETYPE ptr;
   char *s1;

   s1=getparam10(s,&d1);
   if ((s1!=NULL) && (DirFiles>=d1) && (d1>0)) {
     putmsg("\r\n");
     ptr=(LPFILETYPE)(DirTbl+(DWORD)sizeof(FILETYPE)*((DWORD)d1-1UL));
     if  (ptr->flag & 2)
       DispDirectory(ptr->lba,ptr->size,ISJOLIET);
     else 
       putmsg("Selected item is not directory.\r\n");
   }
   else
     dispparamerr(s);
}

void _EXEC_DISPID3(int cmd,char *s)
{
   WORD d1;
   LPFILETYPE ptr;
   char *s1;

   s1=getparam10(s,&d1);
   if ((s1!=NULL) && (DirFiles>=d1) && (d1>0)) {
     putmsg("\r\n");
     ptr=(LPFILETYPE)(DirTbl+(DWORD)sizeof(FILETYPE)*((DWORD)d1-1UL));
     if  ((ptr->flag & 2)==0) {
       if (ptr->ismp3) {
#ifdef __BORLANDC__
	_asm {
		mov ax,12h
		int 10h
	}
	_disable();
	outpw(0x3ce,0x0005); /* writemode 0    */
	outpw(0x3ce,0x0001); /* Latch _disable */
	outpw(0x3ce,0xff08); /* all bit enable */
	outpw(0x3ce,0x0003); /* set mode/rotate=0 */
    outpw(0x3c4,0x0f02); /* sequencer */
    _enable();
#endif
     clearplaylist();
     addplaylist(ptr->lba,ptr->size,ptr->filename);
     mp3play(0,1);
#ifdef __BORLANDC__
	_asm {
	  mov	ax,3h
	  int	10h
	}
#endif
       } else {
         putmsg("Selected item is not mp3 file.\r\n");
       }
     } else 
       putmsg("Selected item is directory.\r\n");
   }
   else
     dispparamerr(s);
}

void SendDspDat(BYTE far *p,DWORD size)
{
#define MAX_SENDSIZE 0xF000
   DWORD ptrn,remainsize;
   WORD i,siz,seg,ofs;
   BYTE dat;

   ptrn=(DWORD)p;
   ptrn=((ptrn >> 12) & 0xFFFF0) + (ptrn & 0xFFFF); /* physical address */
   remainsize=size;
   while (remainsize>0) {
      /* fix pointer */
     seg=ptrn >> 4; ofs=ptrn & 0xf;
     p=(BYTE far *)MK_FP(seg,ofs); /**/
     if (remainsize<=MAX_SENDSIZE) siz=remainsize;
     else siz=MAX_SENDSIZE;
     for (i=0;i<siz;i++) {
       while ((inp(FPGAR2) & DEMAND)==0)  /* WAIT DEMAND */
       ATA_INP(ATA_DATA);
       while (inp(FPGAR2) & SREGEMP)  /* wait empty shiftreg. */
       ATA_INP(ATA_DATA);
       dat=*p++;
       outp(FPGAR3,dat); /*senddac*/
       ATA_INP(ATA_DATA);
     }
     ptrn+=siz;
     remainsize-=siz;
     sprintf(dispbuf,"remain %ld bytes\r\n",remainsize);
     putmsg(dispbuf);
   }
}

void _EXEC_LOADDATA(int cmd,char *s)
{
   WORD d1;
   DWORD siz,rsize,lba;
   LPFILETYPE ptr;
   char *s1;
   int secs;

#ifdef __BORLANDC__
/*   siz=0xffe0UL;*/
   siz=0x1000000UL;
#else
   siz=0x30000UL;
#endif
   s1=getparam10(s,&d1);
   if ((s1!=NULL) && (DirFiles>=d1) && (d1>0)) {
     putmsg("\r\n");
     ptr=(LPFILETYPE)(DirTbl+(DWORD)sizeof(FILETYPE)*((DWORD)d1-1UL));
     Loadedsize=0; DirFiles=0;
     rsize=ptr->size; lba=ptr->lba;
     if  ((ptr->flag & 2) == 0) {
       secs=LoadingData(LocalBuf,siz,lba,rsize);
	   if (rsize>secs*2048) Loadedsize=secs*2048;
	   else Loadedsize=rsize;
     }
     else 
       putmsg("Selected item is directory.\r\n");
   }
   dispparamerr(s);
}

void _EXEC_DIR(int cmd,char *s)
{
   ReadPathTable();
   DELAY(1);
   DispDirectory(RootDir,RootDirLen,ISJOLIET);
}

void _EXEC_DIRD(int cmd,char *s)
{
#ifdef __BORLANDC__
	_asm {
		mov ax,12h
		int 10h
	}
	_disable();
	outpw(0x3ce,0x0005); /* writemode 0    */
	outpw(0x3ce,0x0001); /* Latch _disable */
	outpw(0x3ce,0xff08); /* all bit enable */
	outpw(0x3ce,0x0003); /* set mode/rotate=0 */
    outpw(0x3c4,0x0f02); /* sequencer */
    _enable();
#endif
	Grph_Init();
    mainmenu();
#ifdef __BORLANDC__
	_asm {
	  mov	ax,3h
	  int	10h
	}
#endif
}

void _EXEC_PLAYDATA(int cmd,char *s)
{
   WORD d1;
   DWORD rsize,lba;
   LPFILETYPE ptr;
   char *s1;

   s1=getparam10(s,&d1);
   if ((s1!=NULL) && (DirFiles>=d1) && (d1>0)) {
     putmsg("\r\n");
     ptr=(LPFILETYPE)(DirTbl+(DWORD)sizeof(FILETYPE)*((DWORD)d1-1UL));
     rsize=ptr->size; lba=ptr->lba;
     if  ((ptr->flag & 2) == 0)
       PlayFile(lba,0,rsize);
     else 
       putmsg("Selected item is directory.\r\n");
   }
   dispparamerr(s);
}

void _EXEC_SENDDAC(int cmd,char *s)
{
   WORD d1,d2;
   char *s1;

   s1=getparam(s,&d1);
   if (s1!=NULL) {
      if (getparam(s1,&d2)) {
        if (!SendI2C(0x9a,d1,d2)) putmsg("send ng\r\n");
        sprintf(dispbuf,"write $%X to port $%X\r\n",d2,d1);
        putmsg(dispbuf);
        return;
      }
   } else
     dispparamerr(s);
}


void _EXEC_SENDDSP(int cmd,char *s)
{
   DWORD dat;

   if (GetMemory(0x300,&dat)){ 
     sprintf(dispbuf,"data $%X\r\n",dat);
     putmsg(dispbuf);
     return;
   } else
     putmsg("error!\r\n");
}

void _EXEC_SETFEATURE(int cmd,char *s)
{
   WORD d1;
   char *s1;
   ATAPAC pac;

   clratapac(&pac);
   s1=getparam(s,&d1);
   if (s1!=NULL) {
     pac.cmd=0x3;
     pac.seccnt=d1;
     SendATACommand(&pac);
     putmsg("ok\r\n");
   }
}

void _EXEC_DISPG(int cmd,char *s)
{
#ifndef __BORLANDC__
  far_memcpy(_VRAM_BASE,(BYTE far *)0xE0000000UL,640 / 8 * 200);
#endif
}

void (*monitorfunc[])(int cmd,char *s)=
   {_EXEC_INPORT,_EXEC_INPORT,_EXEC_OUTPORT,_EXEC_OUTPORT,_EXEC_DUMP,_EXEC_FILL,_EXEC_PROG,
   _EXEC_CLEAR,_EXEC_LOAD,_EXEC_EJECT,_EXEC_RESET,_EXEC_READTOC,_EXEC_REQUESTSENSE,_EXEC_READDATA,
   _EXEC_DIR,_EXEC_PLAY,_EXEC_STOP,_EXEC_SENDDAC,_EXEC_CHDIR,_EXEC_LOADDATA,_EXEC_SENDDSP,
   _EXEC_SETFEATURE,_EXEC_PLAYDATA,_EXEC_DISP,_EXEC_DISPG,_EXEC_DIRD,_EXEC_DISPID3};

void monitor(void)
{
   char buf[MAX_INPUT_LEN];
   char inputbuf[MAX_INPUT_LEN+1];
   char *s;
   int cmd;

   _DEFSEG=_DEFOFS=0;  /*seg:ofs=0000:0000*/
   initcomm();
   putmsg("\r\nV53A Monitor program by myan\r\n");
   while (1) {
     putmsg("COMMAND>");
     getmsg(1,inputbuf);  /* input line. */
     strcpy(buf,inputbuf);
     strupr(buf);  /* to large chr. */
     cmd=getcmd(&s,buf);  /* check command */
     if (cmd==_QUIT) break;
     if ((cmd==_NOP) || (cmd==-1)) continue;
     if (cmd==_DISP) _EXEC_DISP(cmd,&inputbuf[s-buf]);
     else            monitorfunc[cmd](cmd,s);
   }
}
