/* MPROM Firmware Code version 1.0 (I2C BUS Control Library)         */
/* programmed by M.yanagisawa(MoroboshiLUM)  elfin@mth.biglobe.ne.jp */
/* Copyright (c) 2000 by M.yanagisawa. All rigths reserved.          */

#ifdef __BORLANDC__

#include <stdio.h>
#include <dos.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <alloc.h>

#else

#include <stdrom.h>

#endif

#include "generic.h"
#include "mprom.h"

void DELAY(int cnt); /* */
void DELAY10MS(int cnt); /* */

/*********** I2C Bus Control Functions ************/

void I2CWait(void)
{
 /*6us wait*/
 int i;
 for (i=0;i<50;i++) inp(FPGAR1); /* 120ns * 50 (UPPER) */
}

void I2CStart(void)
{
  BYTE dat;
  
  dat=inp(FPGAR1);
  outp(FPGAR1,dat=(dat & ~I2C_DATA)); /*SDA = L*/
  I2CWait();
  outp(FPGAR1,dat=(dat |  I2CDOUT)); /*SDA = Out*/
  I2CWait();
  outp(FPGAR1,dat=(dat & ~I2C_COM)); /*SCL = L */
  I2CWait();
}

void I2C2NDStart(void)
{
  BYTE dat;
  
  /* after ACK , DATA = L, CLK = L */
  dat=inp(FPGAR1);
  outp(FPGAR1,dat=(dat | I2C_DATA)); /*SDA = H*/
  I2CWait();
  outp(FPGAR1,dat=(dat | I2CDOUT));  /*SDA = Out*/
  I2CWait();
  outp(FPGAR1,dat=(dat | I2C_COM)); /*SCL = H */
  I2CWait();
  outp(FPGAR1,dat=(dat & ~I2C_DATA)); /*SDA = L*/
  I2CWait();
  outp(FPGAR1,dat=(dat & ~I2C_COM)); /*SCL = L */
  I2CWait();
}

void I2CStop(void)
{
  BYTE dat;
  
  dat=inp(FPGAR1);
  outp(FPGAR1,dat=(dat & ~I2C_DATA)); /*SDA = L*/
  I2CWait();
  outp(FPGAR1,dat=(dat | I2CDOUT));   /*SDA = Out*/
  I2CWait();
  outp(FPGAR1,dat=(dat | I2C_COM));   /*SCL = H */
  I2CWait();
  outp(FPGAR1,dat=(dat | I2C_DATA));  /*SDA = L->H*/
  I2CWait();
}

BOOL I2CSendByte(BYTE c)
{
   BYTE MASK,dat,indat;
   int i;
   
   MASK=0x80;
   for (i=0;i<8;i++) {
     dat=inp(FPGAR1);
     if (MASK & c) outp(FPGAR1,dat=(dat | I2C_DATA));
     else          outp(FPGAR1,dat=(dat & ~I2C_DATA));
     outp(FPGAR1,dat=(dat | I2CDOUT)); /* SDAT = Out */
     I2CWait();
     outp(FPGAR1,dat=(dat | I2C_COM)); /* SCL L->H */
     I2CWait();
     outp(FPGAR1,dat=(dat & ~I2C_COM)); /* SCL H->L */
     I2CWait();
     MASK >>= 1;
   }
   /*RecvACK*/
   dat=inp(FPGAR1);
   outp(FPGAR1,dat=(dat & ~I2CDOUT)); /* In */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2C_COM)); /* SCL L->H */
   I2CWait();
   indat=inp(FPGAR2);
   outp(FPGAR1,dat=(dat & ~I2C_COM)); /* SCL H->L */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2CDOUT));  /* Out */
   I2CWait();
   return((indat & 0x1)==0);
}

void I2CSendACK(void)
{
   WORD dat;

   dat=inp(FPGAR1);
   outp(FPGAR1,dat=(dat & ~I2C_DATA)); /* ACK(L) */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2CDOUT));   /* Out */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2C_COM)); /* SCL L->H */
   I2CWait();
   outp(FPGAR1,dat=(dat & ~I2C_COM)); /* SCL H->L */
   I2CWait();
}

void I2CSendNAK(void)
{
   WORD dat;

   dat=inp(FPGAR1);
   outp(FPGAR1,dat=(dat | I2C_DATA));  /* NAK(H) */
   outp(FPGAR1,dat=(dat | I2CDOUT));   /* Out */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2C_COM));  /* SCL L->H */
   I2CWait();
   outp(FPGAR1,dat=(dat & ~I2C_COM)); /* SCL H->L */
   I2CWait();
   outp(FPGAR1,dat=(dat | I2CDOUT));  /* Out */
   I2CWait();
}

BYTE I2CRecvByte(void)
{
   BYTE dat,indat,r;
   int i;
   
   r=0;
   dat=inp(FPGAR1);
   outp(FPGAR1,dat=(dat & ~I2CDOUT));
   I2CWait();
   for (i=0;i<8;i++) {
     outp(FPGAR1,dat=(dat | I2C_COM)); /* SCL L->H */
     I2CWait();
     indat=inp(FPGAR2);
     if (indat & I2C_DATA) r|=1;
     if (i!=7) r<<=1;
     outp(FPGAR1,dat=(dat & ~I2C_COM)); /* SCL H->L */
     I2CWait();
   }
   return(r);
}

void I2CInit(void)
{
  BYTE dat;
  
  dat=inp(FPGAR1);
  outp(FPGAR1,dat | I2C_DATA | I2C_COM | I2CCOUT | I2CDOUT);
/* I2CC/CD is Output, SDA,SCL = 1*/
}


BOOL SendI2C(BYTE devadr,BYTE sa,BYTE dat)
{
  I2CStart();
  if (!I2CSendByte(devadr)) return(FALSE);
  if (!I2CSendByte(sa)) return(FALSE);
  if (!I2CSendByte(dat)) return(FALSE);
  I2CStop();
  return(TRUE);
}

BOOL SendI2CW(BYTE devadr,BYTE sa,WORD dat)
{
  I2CStart();
  if (!I2CSendByte(devadr)) return(FALSE);
  if (!I2CSendByte(sa)) return(FALSE);
  if (!I2CSendByte((dat >> 8) & 0xff)) return(FALSE);
  if (!I2CSendByte(dat & 0xff)) return(FALSE);
  I2CStop();
  return(TRUE);
}

BOOL RecvI2C(BYTE devadr,BYTE sa,BYTE *dat)
{
  I2CStart();
  if (!I2CSendByte(devadr | 1)) return(FALSE);
  if (!I2CSendByte(sa)) return(FALSE);
  *dat=I2CRecvByte();
  I2CSendACK();
  I2CStop();
  return(TRUE);
}

BOOL RecvI2CW(BYTE devadr,BYTE sa,WORD *dat)
{
  BYTE tmp1,tmp2;
  I2CStart();
  if (!I2CSendByte(devadr | 1)) return(FALSE);
  if (!I2CSendByte(sa)) return(FALSE);
  tmp1=I2CRecvByte();
  I2CSendACK();
  tmp2=I2CRecvByte();
  I2CSendACK();
  I2CStop();
  *dat=((WORD)tmp1 << 8)+(WORD)tmp2;
  return(TRUE);
}

/* set DAC mixer */
void SetDACMute(BOOL isCDDA,BOOL isMP3)
{
   BYTE dat;
   dat=0;
   if (isCDDA) dat|=0x08;
   if (isMP3)  dat|=0x04;
   SendI2C(0x9a,0xc3,dat);
   DELAY(10);
}

/* get played frame count */
BOOL GetMemory(WORD addr,DWORD *data)
{
   WORD tmp[3];
   int i;

  I2CStart();
  if (!I2CSendByte(0x3a)) return(FALSE);
  if (!I2CSendByte(0x68)) return(FALSE);
  if (!I2CSendByte(0xe0)) return(FALSE); /* Read D0 Memory */
  if (!I2CSendByte(0x00)) return(FALSE);
  if (!I2CSendByte(0x00)) return(FALSE); /* words */
  if (!I2CSendByte(0x01)) return(FALSE);
  if (!I2CSendByte((addr & 0xff00) >> 8)) return(FALSE); /* addr */
  if (!I2CSendByte(addr & 0xff)) return(FALSE);
  I2CStop();
  I2CWait();
  /* execute default read command for MAS3507D */
  I2CStart();
  if (!I2CSendByte(0x3a)) return(FALSE); /* device write */
  if (!I2CSendByte(0x69)) return(FALSE); /* data read */
  I2C2NDStart();
  if (!I2CSendByte(0x3b)) return(FALSE); /* device read */
  for (i=0;i<4;i++) {
    tmp[i]=I2CRecvByte();
    I2CSendACK();
  }
  I2CStop();
  *data=((DWORD)tmp[0] << 8) | (DWORD)tmp[1] | ((DWORD)tmp[3] << 16);
  return(TRUE);
}

/* get played frame count */
BOOL GetFrameNumDefault(WORD *framenum)
{
   WORD tmp1,tmp2;

  /* execute default read command for MAS3507D */
  I2CStart();
  if (!I2CSendByte(0x3a)) return(FALSE); /* device write */
  if (!I2CSendByte(0x69)) return(FALSE); /* data read */
  I2C2NDStart();
  if (!I2CSendByte(0x3b)) return(FALSE); /* device read */
  tmp1=I2CRecvByte();
  I2CSendACK();
  tmp2=I2CRecvByte();
  I2CSendNAK();
  I2CStop();
  *framenum=(tmp1 << 8) | tmp2;
  return(TRUE);
}

void SetDACVolume(WORD RVol,WORD LVol,BOOL isEMPH)
{
  WORD dat;

  dat=((LVol & 0x3f) << 8) | (RVol & 0x3f);
  if (isEMPH) dat|=0x4000;
  SendI2CW(0x9a,0xc2,dat);
  DELAY(5);
}

void InitDAC(void)
{
  /* DSP is used inverted 32bit SDO mode. So, LR_SEL is 1  */
  SendI2C(0x9a,0xc1,0x10);
  DELAY(5);
}

/* DSP & DAC Reset & I2C Bus Reset */
void DACDSPReset(void)
{
  BYTE dat;
  
  dat=inp(FPGAR1);
  outp(FPGAR1,dat=(dat & ~DSPEN & ~RESETDSP));
  I2CWait();
  I2CInit();
  dat=inp(FPGAR1);
  outp(FPGAR1,dat | RESETDSP);
  DELAY(2);
  SetDACMute(FALSE,FALSE); /* MP3 & CDDA Mute */
  InitDAC(); /* set DAC WSL mode */
}
