/* MPROM Firmware Code version 1.0                                   */
/* programmed by M.yanagisawa(MoroboshiLUM)  elfin@mth.biglobe.ne.jp */
/* Copyright (c) 2000 by M.yanagisawa. All rigths reserved.          */

#ifdef __BORLANDC__

#include <stdio.h>
#include <dos.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <alloc.h>

#else

#include <stdrom.h>

#endif

#include "generic.h"
#include "drawlib.h"
#include "rslib.h"
#include "mprom.h"
#include "i2clib.h"
#include "main.h"
#include "atactrl.h"
#include "flashlib.h"

char ID3TAGID[]={'T','A','G'};

/* ID3 TAG LIST */
char *ID3TAG_GENRESTR[] = {
	"Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
	"Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
	"Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
	"Death Metal", "Pranks", "Soundtrack", "Euro-tecno", "Ambient", "Trip-Hop",
	"Vocal", "Juzz+Funk", "Fusion", "Trance", "Classical", "Instrumental",
	"Acid", "House", "Game", "Sound Clip", "Gospel", "Noise", "AlternRock",
	"Bass", "Soul", "Punk", "Space", "Meditative", "Instrumental Pop",
	"Instrumental Rock", "Ethnic", "Gothic", "Darkwave", "Techno-Industrial",
	"Electronic", "Pop-Folk", "Eurodance", "Dream", "Southern Rock", "Comedy",
	"Cult", "Gangsta", "Top 40", "Christian Rap", "Pop/Funk", "Jungle",
	"Native American", "Cabaret", "New Wave", "Psychadelic", "Rave",
	"Showtunes", "Trailer", "Lo-Fi", "Tribal", "Acid Punk", "Acid Jazz",
	"Polka", "Retro", "Musical", "Rock & Roll", "Hard Rock", "Folk",
	"Folk/Rock", "National Folk", "Swing", "Fast-Fusion", "Bedob", "Latin",
	"Revival", "Celtic", "Bluegrass", "Avantagarde", "Gothic Rock",
	"Progressive Rock", "Psychadelic Rock", "Symphonic Rock", "Slow Rock",
	"Big band", "Chorus",  "Easy Listening", "Acoustic", "Humour", "Speech",
	"Chanson", "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass",
	"Primus", "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
	"Folklore", "Ballad", "Power Ballad", "Rhytmic Soul", "Freestyle",
	"Duet", "Punk Rock", "Drum Solo", "Acapella", "Euro-House", "Dance Hall",
	"Goa", "Drum & Bass", "Club-House", "Hardcore", "Terror", "Indie",
	"BritPop", "Negerpunk", "Polsk Punk", "Beat", "Christian Gangsta Rap",
	"Heavy Metal", "Black Metal", "Crossover", "Contemporary Christian",
	"Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
	"Synthpop","","", /* 149 */
	"","","","","","","","","","", /* 150-159 */
	"","","","","","","","","","", /* 160-169 */
	"","","","","","","","","","", /* 170-179 */
	"","","","","","","","","","", /* 180-189 */
	"","","","","","","","","","", /* 190-199 */
	"","","","","","","","","","", /* 200-209 */
	"","","","","","","","","","", /* 210-219 */
	"","","","","","","","","","", /* 220-229 */
	"","","","","","","","","","", /* 230-239 */
	"","","","","","","","","Heavy Rock(J)","Doom Rock(J)", /* 240-249 */
	"J-POP(J)","Seiyu(J)","Tecno Ambient(J)","Moemoe(J)","Tokusatsu(J)","Anime(J)"	/* 250-255 */
};

int cs_bitrate[2][3][14] = {
	{
		{ 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448 },
		{ 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384 },
		{ 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320 }
	}, {
		{ 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256 },
		{ 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160 },
		{ 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160 }
	   }
	};

char *cs_mode[] = {
   "Stereo", "Joint stereo", "Dual channel", "Mono" };

char *cs_emphasis[] = {
    "None", "50/15ms", "Reserved", "CCITT J.17"
    };

WORD cs_sfreq[2][3] = {
     { 44100, 48000, 32000 }, { 22050, 24000, 16000 }
};

char dispbuf[256];   /* for display buffer*/

BYTE far *loadbuff[2];
BOOL ENABLEDMA;

BYTE far *PathTbl;  /* ISO9660 Path Table */
BYTE far *DirTbl;
BYTE far *LocalBuf; /* Global Read Buffer(tmp) */
BYTE VOLUMELABEL[34];  /* Volume Label(ISO9660) */
BYTE VOLUMELABELJ[34]; /* Volume Label(Joliet)  */
BYTE VOLUMEDATE[40];   /* Volume TIME  */
DWORD RootLBA;
DWORD RootDir;
DWORD RootDirLen;
int DirFiles;
int TrueDirFiles;
DWORD Loadedsize;

#define PLAYLISTMAX 200

int PLAYLISTCNT;
LPPLAYLIST PlayList; /* Play List */

/* Joliet UCS-2 escape sequense */
BYTE JOLIETESC[3][3] = {
  { 0x25,0x2f,0x40 },
  { 0x25,0x2f,0x43 },
  { 0x25,0x2f,0x45 }
};

LPFILETYPE DIRMAP[MAX_DIRMAP];
BYTE TRMAP[MAX_DIRMAP];

int isOpenNum;
TWinStack WinStack[MAX_OVERLAP];

/* LCD VRAM Steal area */
#ifdef __BORLANDC__

/* for IBM-PC/AT VGA */
FILE *FontFile;
FILE *UnicodeFile;
BYTE LocalKanjiBuf[16*2];
BYTE LocalAnkBuf[16*256];

#endif

int TrackNum;
DWORD TrackLba[100];
BOOL isAudioTrack[100];

WORD KanjiMap[256*4];  /*Kanji */
WORD FontOfs; /* Font Body Offset */
BYTE far *KanjiRom;
BYTE far *AnkRom;
char ISO9660_ID[]={'C','D','0','0','1'};
BOOL ISJOLIET;
BOOL ISEMPH;
WORD MAINVOLUME;
int PLAYINGTRACK;

BOOL ISLOADERROR;
BOOL ISPLAYING;
BOOL ISMP3;
int CURRENTDISCTYPE;
WORD _TIMER_DIV;
BOOL _ENABLE_TIMER;
BOOL _ENABLE_DISP;
TMPEGINFO CURRENTINFO;
long PREVSEC;
DWORD _MP3_STARTTIME;
DWORD _GLOBAL_COUNTER_;
int PLAYINGSTAT;

/* --------------------------- */
/* Graphics Lowlevel functions */
/* --------------------------- */

#ifdef __BORLANDC__
/* unicode conversion table 1 */
unsigned short unitbl1[]= {
     0x0000,0xFFFF,0xFFFF,0x0100,0x0200,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0x0300,0x0400,0x0500,0x0600,0xFFFF,0x0700,0x0800,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0x0900,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0x0A00,0x0B00,
     0x0C00,0x0D00,0x0E00,0x0F00,0x1000,0x1100,0x1200,0x1300,0x1400,0x1500,0x1600,0x1700,0x1800,0x1900,0x1A00,0x1B00,
     0x1C00,0x1D00,0x1E00,0x1F00,0x2000,0x2100,0x2200,0x2300,0x2400,0x2500,0x2600,0x2700,0x2800,0x2900,0x2A00,0x2B00,
     0x2C00,0x2D00,0x2E00,0x2F00,0x3000,0x3100,0x3200,0x3300,0x3400,0x3500,0x3600,0x3700,0x3800,0x3900,0x3A00,0x3B00,
     0x3C00,0x3D00,0x3E00,0x3F00,0x4000,0x4100,0x4200,0x4300,0x4400,0x4500,0x4600,0x4700,0x4800,0x4900,0x4A00,0x4B00,
     0x4C00,0x4D00,0x4E00,0x4F00,0x5000,0x5100,0x5200,0x5300,0x5400,0x5500,0x5600,0x5700,0x5800,0x5900,0x5A00,0x5B00,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
     0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0x5C00
    };

#endif


#ifdef __BORLANDC__
/* LSIC-86 only. */
void far_memset(BYTE far *p,BYTE dat,int size)
{
   int i;
   
   for (i=0;i<size;i++) *p++=dat;
}
#endif

void steal_onoff(BOOL OnOff)
{
   if (OnOff) 
     outp(FPGAR4,inp(FPGAR4) | STEALON);
   else
     outp(FPGAR4,inp(FPGAR4) & ~STEALON);
}

void disp_onoff(BOOL OnOff)
{
   if (OnOff) 
     outp(FPGAR4,inp(FPGAR4) | (DOFF1 | DOFF2));
   else
     outp(FPGAR4,inp(FPGAR4) & ~(DOFF1 | DOFF2));
}

void lcd_onoff(BOOL OnOff)
{
   if (OnOff) 
     outp(FPGAR4,inp(FPGAR4) | (STEALON | LCDON));
   else
     outp(FPGAR4,inp(FPGAR4) & ~(STEALON | LCDON));
}

#ifdef __BORLANDC__
WORD Unicode2SJIS(WORD code)
{
   WORD buf,ofs;

   if (unitbl1[code >> 8]==0xffff) return(0xffff);
   else {
     ofs=unitbl1[code>>8]+(code & 0xff);
     fseek(UnicodeFile,ofs << 1,SEEK_SET);
     fread(&buf,sizeof(buf),1,UnicodeFile);
     if (buf) return(buf);
     else return(0xffff);
   }
}
#else

WORD Unicode2SJIS(WORD code);

#endif

/* key read routine. (wait while any key pushed) */
int readkey(void)
{
   int i,j,result,r;
   WORD MASK,keydata1,keydata2;
#ifdef __BORLANDC__
   int c;

   while(1) {
     c=getch();
     if (isalpha(c))
       if (islower(c)) c=toupper(c);
     switch (c) {
       case 'U' : return(_KEY_OF_UP);
       case 'D' : return(_KEY_OF_DOWN);
       case 'R' : return(_KEY_OF_RIGHT);
       case 'L' : return(_KEY_OF_LEFT);
       case 'F' : return(_KEY_OF_FUNC);
       case 'S' : return(_KEY_OF_STOP);
       case 'Z' : return(_KEY_OF_VOLDOWN);
       case 'A' : return(_KEY_OF_VOLUP);
       case 'X' : return(_KEY_OF_ALL);
       case 0x0d : return(_KEY_OF_EXEC);
       case 0x1b : return(_KEY_OF_ESC);
     }
   }
#else
   /* waiting key-pushed */
   result=r=-1;
   while (1) {
     keydata1=inpw(_8255D_A) & 0xfff;	/* first low-byte is PA, high-byte is PB */
     inp(FPGAR2);			/* dummy */
     inp(FPGAR2);			/* dummy */
     keydata2=inpw(_8255D_A) & 0xfff;	/* second (avoid to key-chatering */
     inp(FPGAR2);			/* dummy */
     inp(FPGAR2);			/* dummy */
     if ((keydata1==keydata2) && (keydata1!=0xfff)) {
       MASK=0x0001;
       j=0;
       /* i=0..11 are same _KEY_OF_* */
       for (i=0;i<12;i++) {
         if ((keydata1 & MASK)==0) {
           r=i;
           j++;
         }
         MASK<<=1;
       }
       /* roll-over check */
       if (j==1) result=r;
     } else
       if ((keydata1==keydata2) && (keydata1==0xfff) && (result!=-1)) return(result);
   }
#endif
}

/* key read routine. (no wait. if not keypressed then return -1 */
int readkeypressed(void)
{
   int i,j,r;
   WORD MASK,keydata1,keydata2;
#ifdef __BORLANDC__
   int c;
   
   if (!kbhit()) return(-1);
   c=getch();
   if (isalpha(c))
     if (islower(c)) c=toupper(c);
   switch (c) {
     case 'U' : return(_KEY_OF_UP);
     case 'D' : return(_KEY_OF_DOWN);
     case 'R' : return(_KEY_OF_RIGHT);
     case 'L' : return(_KEY_OF_LEFT);
     case 'F' : return(_KEY_OF_FUNC);
     case 'S' : return(_KEY_OF_STOP);
     case 'Z' : return(_KEY_OF_VOLDOWN);
     case 'A' : return(_KEY_OF_VOLUP);
     case 'X' : return(_KEY_OF_ALL);
     case 0x0d : return(_KEY_OF_EXEC);
     case 0x1b : return(_KEY_OF_ESC);
   }
   return(-1);
#else
   /* waiting key-pushed */
   keydata1=inpw(_8255D_A) & 0xfff;	/* first low-byte is PA, high-byte is PB */
   inp(FPGAR2);			/* dummy */
   inp(FPGAR2);			/* dummy */
   keydata2=inpw(_8255D_A) & 0xfff;	/* second (avoid to key-chatering */
   inp(FPGAR2);			/* dummy */
   inp(FPGAR2);			/* dummy */
   if (keydata1!=keydata2) return(-1);
   if (keydata1==0xfff) return(-1);
   MASK=0x0001;
   j=0;
   /* i=0..11 are same _KEY_OF_* */
   for (i=0;i<12;i++) {
     if ((keydata1 & MASK)==0) {
       r=i;
       j++;
     }
     MASK<<=1;
   }
   /* roll-over check */
   if (j==1) return(r);
   else return(-1);
#endif
}

/* key read routine. (no wait. if not keypressed then return -1 */
void waitnotkeypressed(void)
{
   WORD keydata1,keydata2;
#ifdef __BORLANDC__
   while (kbhit());
#else
   /* waiting key-pushed */
   while (1) {
     keydata1=inpw(_8255D_A) & 0xfff;	/* first low-byte is PA, high-byte is PB */
     inp(FPGAR2);			/* dummy */
     inp(FPGAR2);			/* dummy */
     keydata2=inpw(_8255D_A) & 0xfff;	/* second (avoid to key-chatering */
     inp(FPGAR2);			/* dummy */
     inp(FPGAR2);			/* dummy */
   
     if ((keydata1==keydata2) && (keydata1==0xfff)) break;
   }
#endif
}

/* for FONTX2 format Font file */
long GetFontOfs(WORD code)
{
   WORD *map;
   WORD st,en;
   DWORD ofs;
   int i;
   
   ofs=0;
   map=(WORD *)KanjiMap;
   for (i=0;i<256;i++) {
     st=*map++; en=*map++;
     if (!st) return(-1);
     if ((st<=code) && (en>=code)) return((ofs+code-st) * KFONT_SIZE);
     ofs+=((en-st)+1);
   }
   return(-1);
}

BYTE far *GetZFont(WORD code)
{
  long ofs;
  
  ofs=GetFontOfs(code);
  if (ofs<0) return(NULL);
#ifdef __BORLANDC__
  fseek(FontFile,ofs+FontOfs,SEEK_SET);
  fread(LocalKanjiBuf,KFONT_SIZE,1,FontFile);
  return((BYTE far *)LocalKanjiBuf);
#else
  return(KanjiRom+ofs+FontOfs); /* LSIC-86 always make far-pointer regulared! */
#endif
}

BYTE far *getvram(int cx,int cy)
{
   return(((cx >> 3) + cy*80) + _VRAM_BASE);
}

/* displaying kanji/ank characters. x,y is dot axis. */
void putkanji(int x,int y,WORD code)
{
   BYTE far *font;
   
   font=NULL;
   if (code>=0x100) {
     font=GetZFont(code);
     if (font)
       Grph_PutImage(x,y,16,16,_WRITEMODE_RPL,font);
   } else {
#ifdef __BORLANDC__
     font=&LocalAnkBuf[AFONT_SIZE*code];
#else
     font=&AnkRom[AFONT_SIZE*code];
#endif
     Grph_PutImage(x,y,8,16,_WRITEMODE_RPL,font);
   }
}

/* displaying messages */
void putmessage(int cx,int cy,BYTE far *s)
{
  WORD isk1;
  
  isk1=0;
  while (*s) {
    if (isk1) {
      isk1|=*s;
      putkanji(cx-CHR_WIDTH,cy,isk1);
      isk1=0;
    } else
      if (((*s>=0x81) && (*s<=0x9f))
         || ((*s>=0xe0) && (*s<=0xfc))) isk1=(WORD)*s << 8;
    else
      putkanji(cx,cy,*s);
    s++;  cx+=CHR_WIDTH;
    if (cx>=SCR_WIDTH) {
      cx=0;
      cy+=RAW_SIZE;
    }
    if (cy>=SCR_HEIGHT) cy-=RAW_SIZE;
  }
}

void vram_clear(void)
{
   far_memset(_VRAM_BASE,0,_VRAM_SIZE);
}

/* drawing window (3D looked like :) */
void draw_window(int x1,int y1,int x2,int y2)
{
   Grph_Box(x1,y1,x2-1,y2-1,_WRITEMODE_RPL,0,0xffff);
   Grph_Line(x1+1,y2,x2,y2,_WRITEMODE_RPL,0,0xffff);
   Grph_Line(x2,y1+1,x2,y2,_WRITEMODE_RPL,0,0xffff);
}

void draw_window_s(int x1,int y1,int x2,int y2)
{
   draw_window(x1+1,y1+1,x2-1,y2-1);
}

/* background */
void dispbackground(void)
{
   vram_clear();
   draw_window(0,0,SCR_WIDTH-1,SCR_HEIGHT-1);
   /**/
   Grph_Line(3,MAX_DISPLINE*RAW_SIZE + 3,
             SCR_WIDTH-4,MAX_DISPLINE*RAW_SIZE + 3,_WRITEMODE_RPL,0,0xffff);
}

/* background(audio mode) */
void dispbackground_audio(void)
{
   vram_clear();
   draw_window(0,0,SCR_WIDTH-1,SCR_HEIGHT-1);
   /**/
   Grph_Line(3,MAX_DISPLINE_AUDIO*RAW_SIZE + 3,
             SCR_WIDTH-4,MAX_DISPLINE_AUDIO*RAW_SIZE + 3,_WRITEMODE_RPL,0,0xffff);
}

/* for open overlap window(2level only) */
void openwindow(int x1,int y1,int x2,int y2)
{
  int n;

  if (isOpenNum<MAX_OVERLAP) {
    if (x1>x2) {
      n=x2; x2=x1; x1=n;
    }
    if (y1>y2) {
      n=y2; y2=y1; y1=n;
    }
    WinStack[isOpenNum].x=x1;
    WinStack[isOpenNum].y=y1;
    WinStack[isOpenNum].width=x2-x1+1;
    WinStack[isOpenNum].height=y2-y1+1;
    Grph_GetImage(x1,y1,x2-x1+1,y2-y1+1,WinStack[isOpenNum].buf); /* backup */
    Grph_FillArea(x1,y1,x2,y2,_WRITEMODE_RPL,0);
    isOpenNum++;
  }
}

/* close window (stack) */
void closewindow(void)
{
  if (isOpenNum) {
    isOpenNum--;
    Grph_PutImage(WinStack[isOpenNum].x,WinStack[isOpenNum].y,
                  WinStack[isOpenNum].width,WinStack[isOpenNum].height,_WRITEMODE_RPL,WinStack[isOpenNum].buf);
  }
}

/* divide sjis code string(wordrap enable) */
void DivStr(BYTE *s,BYTE *d1,BYTE *d2)
{
   int i,smax,divmax,prev;
   int isk1;
   char *orgs;
   
   smax=strlen(s); divmax=smax >> 1; orgs=s;
   i=0; isk1=0; *d1=0; *d2=0; prev=-1;
   for (i=0;i<smax;i++) {
     if (!*s) break;
     if (isk1==1) isk1++;
     else
     if (((*s>=0x81) && (*s<=0x9f))
         || ((*s>=0xe0) && (*s<=0xfc))) isk1=1;
     if (!isk1 && (*s==0x20)) prev=i; /* ank space check */
     if (!isk1 && (divmax<i)) break;
     s++;
     if (isk1==2) isk1=0;
   }
   if (i==smax) {
     strcpy(d1,orgs);
     return;
   }
   if (prev==-1) prev=i;
   if ((prev>divmax-5) && (prev<divmax+5)) ;
   else prev=i;
   s=orgs;
   for (i=0;i<prev;i++) *d1++=*s++; *d1=0;
   for (i=prev;i<=smax;i++) *d2++=*s++; *d2=0;
}

/* cutting string (to avoid dividing sjis-two-byte-code) */
void CutStr(int max,BYTE *s)
{
   int i;
   int isk1;
   
   i=0; isk1=0;
   for (i=0;i<max;i++) {
     if (!*s) break;
     if (isk1==1) isk1++;
     else
     if (((*s>=0x81) && (*s<=0x9f))
         || ((*s>=0xe0) && (*s<=0xfc))) isk1=1;
     s++;
     if (isk1==2) isk1=0;
   }
   if (i==max) {
     if (isk1==1) --s;
     *s=0;
   }
}

#define VOLWIN_X1 420
#define VOLWIN_Y1 120
#define VOLWIN_X2 630
#define VOLWIN_Y2 159
#define VOLWIN_WIDTH (VOLWIN_X2-VOLWIN_X1+1)
#define VOLWIN_HEIGHT (VOLWIN_Y2-VOLWIN_Y1+1)

#define VOLBOX_X1 (VOLWIN_X1+10)
#define VOLBOX_Y1 (VOLWIN_Y1+_MSG_YOFS+RAW_SIZE+2)
#define VOLBOX_X2 (VOLWIN_X2-10)
#define VOLBOX_Y2 (VOLBOX_Y1+8)
#define VOLBOX_WIDTH (VOLBOX_X2-VOLBOX_X1+1)
#define VOLBOX_HEIGHT (VOLBOX_Y2-VOLBOX_Y1+1)

void _setvolume(int volume)
{
#ifndef __BORLANDC__
   _ENABLE_TIMER=FALSE; /* ֎~! */
   if (volume<0) volume=0;
   if (volume>0x38) volume=0x38;
   SetDACVolume(volume,volume,ISEMPH);
   _ENABLE_TIMER=TRUE;
#endif
}

void setvolume(int volume)
{
   int size,vol;
   DWORD lvf;

 /* 1 = -75db , 0x2c = 0db , 0x38 = +18dB */
   if (volume>0x38) volume=0x38;
   vol=volume;
   if (volume<0)  volume=MAINVOLUME;
   if (volume>=0) {
     lvf= volume * 100 / 0x38;
     size=(VOLBOX_WIDTH-2)*lvf/100;
   }
   if (vol<0) {
     Grph_FillArea(VOLBOX_X1+1,VOLBOX_Y1+1,VOLBOX_X2-1,VOLBOX_Y2-1,_WRITEMODE_RPL,0);
     Grph_FillArea(VOLBOX_X1+1,VOLBOX_Y1+1,VOLBOX_X1+size+1,VOLBOX_Y2-1,_WRITEMODE_RPL,5);
   } else if (MAINVOLUME==volume) return;
   else   if (MAINVOLUME>volume) {
     Grph_FillArea(VOLBOX_X1+size+1,VOLBOX_Y1+1,VOLBOX_X2-1,VOLBOX_Y2-1,_WRITEMODE_RPL,0);
   } else {
     Grph_FillArea(VOLBOX_X1+1,VOLBOX_Y1+1,VOLBOX_X1+size+1,VOLBOX_Y2-1,_WRITEMODE_RPL,5);
   }
   _setvolume(volume);
   MAINVOLUME=volume;
}

void openvolume(void)
{
   openwindow(VOLWIN_X1,VOLWIN_Y1,VOLWIN_X2,VOLWIN_Y2);
   draw_window_s(VOLWIN_X1,VOLWIN_Y1,VOLWIN_X2,VOLWIN_Y2);
   putmessage(VOLWIN_X1+_MSG_XOFS,VOLWIN_Y1+_MSG_YOFS,"y{[ݒz");
   Grph_Box(VOLBOX_X1,VOLBOX_Y1,VOLBOX_X2,VOLBOX_Y2,_WRITEMODE_RPL,0,0xffff);
   setvolume(-1);
}

char VOLUMEID[]={'V','O','L'};

void SaveVolume(BYTE vol)
{
   TFLASHBLOCK buf;
   int i;

   memset(buf,0xff,sizeof(buf));
   buf[0]=vol;
   buf[1]=~buf[0]; /* for check */
   for (i=0;i<sizeof(VOLUMEID);i++) buf[i+2]=VOLUMEID[i]; /*for check*/
   Flash_Savedata(&buf);
}

BYTE LoadVolume(void)
{
   TFLASHBLOCK buf;
   int i;
   BYTE vol;

   if (Flash_Loaddata(&buf)) {
     vol=buf[0];
     if ((buf[1]==~vol) && (vol<=0x38)){
       for (i=0;i<sizeof(VOLUMEID);i++) 
         if (buf[i+2]!=VOLUMEID[i]) break;
       if (i==sizeof(VOLUMEID)) return(vol);
     }
   }
   return(0x2c); /* default 0dB */
}

/* must set null-terminate ver. */
void strncpy_far(char far *dest,char far *src,int max)
{
  int i;

  for (i=0;i<max;i++) {
    *dest=*src;
    if (!*src) break;
    dest++; src++;
  }
  *dest=0;
}

LPPLAYLIST getplaylistptr(int index)
{
  return((LPPLAYLIST)((BYTE far *)PlayList+sizeof(TPLAYLIST)*index));
}

int addplaylist(DWORD lba,DWORD size,char far *s)
{
   LPPLAYLIST ptr;

   if (PLAYLISTCNT<PLAYLISTMAX) {
     ptr=getplaylistptr(PLAYLISTCNT);
     ptr->lba=lba;  ptr->size=size;
     strncpy_far(ptr->filename,s,23);
     return(PLAYLISTCNT++);
   } else
     return(-1);
}

int getindex(DWORD lba,DWORD size)
{
   int i;
   LPPLAYLIST ptr;

   ptr=(LPPLAYLIST)PlayList;
   for (i=0;i<PLAYLISTCNT;i++) {
     if ((ptr->lba==lba) && (ptr->size==size)) return(i);
     ptr++;
   }
   return(-1);
}

void deleteplaylist(int index)
{
   LPPLAYLIST ptr,next;
   int n,i;
   
   if (index>PLAYLISTCNT) return;
   ptr=getplaylistptr(index);
   next=ptr+1;
   n=PLAYLISTCNT-index-1;
   for (i=0;i<n;i++) *ptr++=*next++;
   PLAYLISTCNT--;
}

BOOL getplaylist(int index,LPPLAYLIST *ptr)
{
   if (index>PLAYLISTCNT) return(FALSE);
   *ptr=getplaylistptr(index);
   return(TRUE);
}

void clearplaylist(void)
{
  PLAYLISTCNT=0;
}

void drawbtn(int x,int y,BYTE *s)
{
   int ofs,len;

   if (strlen(s)*8<BTN_WIDTH) {
     ofs=(BTN_WIDTH-1-strlen(s)*8) >> 1;
     draw_window(x,y,x+BTN_WIDTH-1,y+BTN_HEIGHT-1);
   } else {
     len=strlen(s)*8+8;
     ofs=4;
     draw_window(x,y,x+len-1,y+BTN_HEIGHT-1);
   }
   putmessage(x+ofs,y+2,s);
}

void flashbtn(int x,int y,BYTE *s)
{
   int i,len;
   
   if (strlen(s)*8>BTN_WIDTH)
     len=strlen(s)*8+8;
   else
     len=BTN_WIDTH;
   for (i=0;i<3;i++) {
     Grph_FillArea(x+1,y+1,x+len-1,y+BTN_HEIGHT-1,_WRITEMODE_XOR,10);
     DELAY10MS(6);
     Grph_FillArea(x+1,y+1,x+len-1,y+BTN_HEIGHT-1,_WRITEMODE_XOR,10);
     DELAY10MS(6);
   }
}

/* multi-line support (max 2lines) */
BOOL messagedlg(char *msg,int mode)
{
   int width,height,lines,x,y,c;
   int b1x,b1y,b2x,b2y;
   char msg1[80];
   char msg2[80];

   if (strlen(msg)>150) CutStr(150,msg);
   width=strlen(msg)*8;
   if (width>MAX_MSGWIDTH) {
     DivStr(msg,msg1,msg2);
     lines=2;
   } else
     lines=1;
   if (lines==1) {
     width=strlen(msg)*8+_MSG_XOFS*2;
   } else {
     if (strlen(msg1)>strlen(msg2)) width=strlen(msg1)*8+_MSG_XOFS*2;
     else                           width=strlen(msg2)*8+_MSG_XOFS*2;
   }
   height=lines*RAW_SIZE+_MSG_YOFS*2;
   if ((mode!=_MSGDLG_NOBTN) && (mode!=_MSGDLG_NOBTN_NOCLOSE)) {
     height=height+BTN_HEIGHT+_MSG_YBTNOFS;
     if (mode==_MSGDLG_1BTN) {
       if (width<BTN_WIDTH+_MSG_XOFS*2+2) width=BTN_WIDTH+_MSG_XOFS*2+2;
     } else {
       if (width<BTN_WIDTH*2+_MSG_XBTNOFS+_MSG_XOFS*2+2) width=BTN_WIDTH*2+_MSG_XBTNOFS+_MSG_XOFS*2+2;
     }
   }
   /* width,height ok */
   x=(SCR_WIDTH-width) >> 1;
   y=(SCR_HEIGHT-height) >> 1;
   /* EXECUTE DRAW */
   openwindow(x,y,x+width-1,y+height-1);
   draw_window_s(x,y,x+width-1,y+height-1);
   if (lines==1) 
     putmessage(x+_MSG_XOFS,y+_MSG_YOFS,msg);
   else {
     putmessage(x+_MSG_XOFS,y+_MSG_YOFS,msg1);
     putmessage(x+_MSG_XOFS,y+_MSG_YOFS+RAW_SIZE,msg2);
   }
   b1x=x+width-1-_MSG_XOFS-BTN_WIDTH;
   b1y=y+height-1-_MSG_YOFS-BTN_HEIGHT;
   b2x=b1x-BTN_WIDTH-_MSG_XBTNOFS;
   b2y=b1y;
   switch (mode) {
     case _MSGDLG_NOBTN_NOCLOSE :
     case _MSGDLG_NOBTN :  break;
     case _MSGDLG_1BTN :
                          drawbtn(b1x,b1y,"OK");
                          break;
     case _MSGDLG_2BTN :
                          drawbtn(b1x,b1y,"ݾ");
                          drawbtn(b2x,b2y,"s");
                          break;
   }
   if ((mode!=_MSGDLG_NOBTN) && (mode!=_MSGDLG_NOBTN_NOCLOSE)) {
     while (1) {
       c=readkey();
       if (c==_KEY_OF_EXEC) {
         if (mode==_MSGDLG_1BTN) flashbtn(b1x,b1y,"");
         else				     flashbtn(b2x,b2y,"");
         break;
       }
       if ((c==_KEY_OF_ESC) && (mode==_MSGDLG_2BTN)) {
         flashbtn(b1x,b1y,"");
         break;
       }
     }
   } else {
     if (mode==_MSGDLG_NOBTN) {
#ifdef __BORLANDC__
       DELAY(300);
#else
	   DELAY10MS(150); /* 1.5s */
#endif
     }
   }
   if (mode!=_MSGDLG_NOBTN_NOCLOSE) closewindow();
   if (mode!=_MSGDLG_2BTN) return(TRUE);
   else if (c==_KEY_OF_ESC) return(FALSE);
   else return(TRUE);
}

int getid3itemlength(int max,char far *p)
{
  int i;
  
  i=max;
  do {
    if ((*p==0) || (*p==' ')) i--;
    else break;
    p--;
  } while (i>0);
  return(i);
}

/* loading id3 tag data */
BOOL LoadID3Tag(DWORD lba,DWORD size,BYTE far *buf)
{
   DWORD lastlba;
   WORD tagofs;
   BYTE far *p;
   BYTE far *org;
   int i;

   lastlba=(size-1) >> 11;  /**/
   if (lastlba>0) lastlba--;
   if (!LoadingData(LocalBuf,0x1000,lba+lastlba,0x1000)) return(FALSE);
   if (size>2048)
     tagofs=(size & 0x7ff)-0x80+0x800;
   else
     tagofs=(size & 0x7ff)-0x80;
   p=LocalBuf; p=p+tagofs; org=p;
   for (i=0;i<0x80;i++) *buf++=*p++;
   for (i=0;i<sizeof(ID3TAGID);i++)
     if (ID3TAGID[i]!=*org) return(FALSE);
     else org++;
   return(TRUE);
}

DWORD getbits(int n,DWORD *buf)
{
  int i;
  DWORD retval;

  retval=0;
  for (i=0;i<n;i++) {
    retval <<= 1;
    if (*buf & 0x80000000UL) retval++;
    *buf <<= 1;
  }
  return(retval);
}

BOOL CheckMP3Info(DWORD info,DWORD len,LPMPEGINFO minfo)
{
  int n;
  DWORD buf;

  buf=(info << 24) | ((info & 0x0000ff00UL) << 8)
      | ((info & 0x00ff0000UL) >> 8) | (info  >> 24);
  if (getbits(12,&buf)==0xFFF) {
    minfo->version = getbits(1,&buf);
	minfo->layer = getbits(2,&buf );
	minfo->errorprotection = getbits(1,&buf);
	minfo->bitrate_index = getbits(4,&buf);
	minfo->sfreq_index = getbits(2,&buf);
	minfo->padding = getbits(1,&buf);
	minfo->extention = getbits(1,&buf);
	minfo->mode_index = getbits(2,&buf);
	minfo->modeextention = getbits(2,&buf);
	minfo->copyright = (BOOL)getbits(1,&buf);
	minfo->original = (BOOL)getbits(1,&buf);
	minfo->emphasis_index = getbits(2,&buf);
    /* check parameters */
    minfo->version = (minfo->version) ? 1:2;
    minfo->errorprotection^=1;
    if (!minfo->layer)
      return(FALSE);
    else
      minfo->layer=4-minfo->layer;
    minfo->bitrate_index--;
    if (minfo->sfreq_index==3) return(FALSE);
    if ((minfo->bitrate_index<0) || (minfo->bitrate_index>=14)) return(FALSE);
    if (minfo->layer==1)
      n=3;
    else
      n=1;
    minfo->frames=(len * 8) / (cs_bitrate[minfo->version-1][minfo->layer-1][minfo->bitrate_index] / n);
    minfo->playbacktime=len / (cs_bitrate[minfo->version-1][minfo->layer-1][minfo->bitrate_index] * 1000 / 8);
    return(TRUE);
  } else
    return(FALSE);
}

/* */
void disptime(BOOL isSign,int x,int y,long sec)
{
   int mm,ss,pmm,pss;
   int c1,c2,xx,yy;
   long pvs;
   void far *body;
   BOOL isMinus;

   if (PREVSEC==sec) return;
   pvs=PREVSEC; /* update */
   PREVSEC=sec; /* update */

   isMinus=FALSE;
   if (sec<0) {
     isMinus=TRUE;
     sec=-sec;
   }
   mm=sec / 60;
   ss=sec - (mm * 60);
   pmm=pvs / 60;
   pss=pvs - (pmm * 60);

   if (isSign) {
     Grph_GetImageData(11,&xx,&yy,&body); /* minus */
     if (((pvs<0) && !isMinus) || (pvs<-65535))
       Grph_FillArea(x,y,x+xx-1,y+yy-1,_WRITEMODE_RPL,0);
     if (isMinus && ((pvs>=0) || (pvs<-65535)))
       Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
     xx++; x+=xx;
   }

   c1=mm / 10;
   c2=pmm / 10;
   Grph_GetImageData(c1,&xx,&yy,&body);
   if ((c1!=c2) && (!c1)) 
     Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
   if (!c1)
     Grph_FillArea(x,y,x+xx-1,y+yy-1,_WRITEMODE_RPL,0);
   xx++; x+=xx;

   c1=mm-(c1 * 10);
   c2=pmm-(c2 * 10);
   Grph_GetImageData(c1,&xx,&yy,&body);
   if (((c1!=c2) && (mm!=0)) || (pvs<-65535))
     Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
   xx++; x+=xx;

   Grph_GetImageData(10,&xx,&yy,&body); /* colon */
   if (pvs<-65535)
     Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
   xx++; x+=xx;

   c1=ss / 10;
   c2=pss / 10;
   Grph_GetImageData(c1,&xx,&yy,&body);
   if ((c1!=c2) || (pvs<-65535))
     Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
   xx++; x+=xx;

   c1=ss-(c1 * 10);
   c2=pss-(c2 * 10);
   Grph_GetImageData(c1,&xx,&yy,&body);
   if ((c1!=c2) || (pvs<-65535))
     Grph_PutImage(x,y,xx,yy,_WRITEMODE_RPL,body);
}

void disptag(BOOL isTagOK,char *msg,LPID3TAG buf)
{
   int xc,x,y,ofs,len,i;

   y=MP3INFO_Y1 + _MSG_YOFS + RAW_SIZE + 3;
   xc = MP3INFO_X2 - MP3INFO_TAGWIDTH + 4; x = xc + 8;
   if (isTagOK) {
     /**/
     putmessage(xc,y,"y^Cgz"); y+=RAW_SIZE;
     len=getid3itemlength(30,(char far *)&buf->title[29]);
     ofs=strlen(msg);
     for (i=0;i<len;i++) msg[i]=buf->title[i];  msg[len]=0;
     Grph_FillArea(x,y,MP3INFO_X2-3,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
     putmessage(x,y,msg);  y+=RAW_SIZE;
     /**/
     putmessage(xc,y,"yA[eBXgz"); y+=RAW_SIZE;
     len=getid3itemlength(30,(char far *)&buf->artist[29]);
     for (i=0;i<len;i++) msg[i]=buf->artist[i];  msg[len]=0;
     Grph_FillArea(x,y,MP3INFO_X2-3,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
     putmessage(x,y,msg);  y+=RAW_SIZE;
     /**/
     putmessage(xc,y,"yAoz"); y+=RAW_SIZE;
     len=getid3itemlength(30,(char far *)&buf->album[29]);
     for (i=0;i<len;i++) msg[i]=buf->album[i];  msg[len]=0;
     Grph_FillArea(x,y,MP3INFO_X2-3,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
     putmessage(x,y,msg);  y+=RAW_SIZE;
     /**/
     strcpy(msg,"yNz"); ofs=strlen(msg);
     len=getid3itemlength(4,(char far *)&buf->year[3]);
     for (i=ofs;i<ofs+len;i++) msg[i]=buf->year[i-ofs];  msg[ofs+len]=0;
     strcat(msg,"N");
     Grph_FillArea(xc,y,MP3INFO_X2-3,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
     putmessage(xc,y,msg);
     /**/
     strcpy(msg,"yެفz");
     strcat(msg,ID3TAG_GENRESTR[buf->genre]);
     CutStr(16,msg);
     putmessage(xc+(MP3INFO_TAGWIDTH >> 1)-8,y,msg);  y+=RAW_SIZE;
     /**/
     putmessage(xc,y,"yRgz"); y+=RAW_SIZE;
     len=getid3itemlength(30,(char far *)&buf->comment[29]);
     for (i=0;i<len;i++) msg[i]=buf->comment[i];  msg[len]=0;
     Grph_FillArea(x,y,MP3INFO_X2-3,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
     putmessage(x,y,msg);
   } else {
     /* tag none */
     Grph_FillArea(xc,y,MP3INFO_X2-3,MP3INFO_Y2-3,_WRITEMODE_RPL,0);
     putmessage(MP3INFO_X2-(MP3INFO_TAGWIDTH >> 1)-44,MP3INFO_Y1+(MP3INFO_HEIGHT >> 1)+_MSG_YOFS,"yȂz");
   }
}

void dispfileinfo(BOOL isOK,char *msg,char far *fname,LPMPEGINFO info)
{
   int x,y;

   x=MP3INFO_X1+_MSG_XOFS; y=MP3INFO_Y1+_MSG_YOFS+RAW_SIZE;
   strncpy_far(msg,fname,40); CutStr(38,msg);
   Grph_FillArea(x+16,y,MP3INFO_X2-MP3INFO_TAGWIDTH-8,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
   putmessage(x+16,y,msg);
   y=MP3INFO_Y2-_MSG_YOFS-BTN_HEIGHT - RAW_SIZE * 2 - 2;
   Grph_FillArea(x,y,MP3INFO_X2-MP3INFO_TAGWIDTH-8,y+RAW_SIZE-1,_WRITEMODE_RPL,0);
   if (isOK) {
   sprintf(msg,"MPEG%d layer-%d %uHz %dKbit/s",
                info->version,
                info->layer,
                cs_sfreq[info->version-1][info->sfreq_index],
                cs_bitrate[info->version-1][info->layer-1][info->bitrate_index]);
   } else
     strcpy(msg,"UNKNOWN FILETYPE");
   putmessage(x,y,msg); y+=RAW_SIZE;
   if (isOK) {
     sprintf(msg,"EMPH = %s / MODE = %s",
                cs_emphasis[info->emphasis_index],
                cs_mode[info->mode_index]);
     putmessage(x,y,msg);
   }
}

const char RIFFID[]={'R','I','F','F'};
const char WAVEID[]={'W','A','V','E'};
const char RMPID[] ={'R','M','P','3'};
const char FMTID[] ={'f','m','t',' '};
const char DATAID[]={'d','a','t','a'};

/* WAVE or RMP */
BOOL FindDataChunk(BYTE far *st,BYTE far **hed,DWORD *size)
{
   BOOL isRIFFOK,isWAVE;
   int j,len;
   BYTE far *p;
   WORD chunksize,fmttag;

   st+=8; /* next RIFF */
   *size=0; isRIFFOK=isWAVE=FALSE; /**/
   /* Check 'WAVE' */
   for (j=0,p=st;j<sizeof(WAVEID);j++)
     if (*p++!=WAVEID[j]) break;
   if (j!=sizeof(WAVEID)) {
     /* Check 'RMP3' */
     for (j=0,p=st;j<sizeof(RMPID);j++)
       if (*p++!=RMPID[j]) break;
     if (j==sizeof(RMPID)) isRIFFOK=TRUE; /*RMP3*/
   } else {
     /*check mp3-formattag(ng-tag?)  WAVE */
     isRIFFOK=isWAVE=TRUE;
   }
   if (isRIFFOK) {
     len=0; st=p;
     while (len<1024) {
       /* check 'fmt ' chunk */
       for (j=0,p=st;j<sizeof(FMTID);j++)
         if (*p++!=FMTID[j]) break;
       if (j==sizeof(FMTID)) {
         fmttag=*(WORD far *)(p+4);
         if ((fmttag!=0x55) && isWAVE) return(FALSE);
       }
       /* check 'data' chunk */
       for (j=0,p=st;j<sizeof(DATAID);j++)
         if (*p++!=DATAID[j]) break;
       if (j==sizeof(DATAID)) {
         *size=*(DWORD far *)p;
         *hed=p+4;
         return(TRUE);
       }
       st+=4; /*skip chunk id */
       chunksize=*(DWORD far *)st+4;
       st+=chunksize; /* next */
       len=len+chunksize+4;
     }
   }
   return(FALSE);
}

/* RIFF/WAVE,RIFF/MP3 CHUNK Analize */
DWORD GetMP3Ofs(BYTE far *st,DWORD *size,BOOL *isriff,BOOL *isok)
{
   BOOL rslt;
   BYTE far *p;
   BYTE far *org;
   int i,j;

   /* Check RIFF */
   org=st; *isriff=*isok=FALSE;
   for (j=0,p=st;j<sizeof(RIFFID);j++)
     if (*p++!=RIFFID[j]) break;
   if (j==sizeof(RIFFID)) {
     /* FOUND! */
     *isriff=TRUE;
     rslt=FindDataChunk(st,&p,size);
     if (rslt) 
       org=p;
     else {
       *isok=FALSE;
       return(0); /* if unknown RIFF then ERROR */
     }
   }
   /* Find $FFFx Data */
   p=org;
   for (i=0;i<256;i++) {
     if (CheckMP3Info(*(DWORD far *)p,0,&CURRENTINFO)) {
       *isok=TRUE;
       *size=*size-(p-st);
       return((DWORD)(p-st));
     }
      else p++;
   }
   *size=0;
   return(0);
}

/* display mp3 file info & playing file status */
int mp3play(int index,int playcnt)
{
   int x,y,c,sec;
   int b1x,b1y,ofs;
   DWORD lba,size,psize;
   char msg[44];
   TID3TAG buf;
   BOOL isTagOK;
   BOOL isVolumeChanging;
   DWORD stVolTimeout;
   int cnt,RESULT;
   LPPLAYLIST playptr;
   int PREVVOLUME;
   BOOL isOK,isriff;
   
   PREVVOLUME = MAINVOLUME;
   RESULT=0;  /* no error */
   openwindow(MP3INFO_X1,MP3INFO_Y1,MP3INFO_X2,MP3INFO_Y2);
   draw_window(MP3INFO_X1,MP3INFO_Y1,MP3INFO_X2,MP3INFO_Y2);

   putmessage(MP3INFO_X2-MP3INFO_TAGWIDTH+4,MP3INFO_Y1+_MSG_YOFS,"yhcR^Oz");
   x=MP3INFO_X2 - MP3INFO_TAGWIDTH + 2;
   y=MP3INFO_Y1 + _MSG_YOFS + RAW_SIZE + 1;
   Grph_Line( x + 1, y, x + MP3INFO_TAGWIDTH - 8,y,_WRITEMODE_RPL,0,0xffff);
   Grph_Line( x - 3, MP3INFO_Y1 + 2 , x - 3 , MP3INFO_Y2- 4 ,_WRITEMODE_RPL,0,0xaaaa);

   x=MP3INFO_X1+_MSG_XOFS; y=MP3INFO_Y1+_MSG_YOFS;
   putmessage(x,y,"yt̃t@Cz");
   y= y + RAW_SIZE * 2;
   Grph_Line( x - 2, y ,MP3INFO_X2 - MP3INFO_TAGWIDTH - 8, y ,_WRITEMODE_RPL,0,0xffff);
   y=MP3INFO_Y2-_MSG_YOFS-BTN_HEIGHT - RAW_SIZE * 2 - 2;
   Grph_Line( x - 2, y-2 ,MP3INFO_X2 - MP3INFO_TAGWIDTH - 8, y-2 ,_WRITEMODE_RPL,0,0xaaaa);
   b1x=MP3INFO_X1+_MSG_XOFS;
   b1y=MP3INFO_Y2-_MSG_YOFS-BTN_HEIGHT;
   drawbtn(b1x,b1y,"tf");

   if (playcnt<0) playcnt=PLAYLISTCNT-index;
   while (1) {
     if (!getplaylist(index,&playptr)) break;
     lba=playptr->lba; size=playptr->size;
     PREVSEC=-65536; sec=0;
     isVolumeChanging=FALSE; ISPLAYING=FALSE;
     if (!isDeviceOK()) {
       RESULT=_REQUEST_CHANGEDISC;
       break;  /* device in ? */
     }
     if (!LoadingData(LocalBuf,0x800,lba,0x800)) {
       RESULT=_REQUEST_CHANGEDISC;
       break; /* load error */
     }
     psize=size;
     ofs=GetMP3Ofs(LocalBuf,&size,&isriff,&isOK);
     if (!isOK) {
       isOK=messagedlg("MP3t@Cł͂܂B",_MSGDLG_1BTN);
     } else {
       /* ---------- */
       isTagOK=LoadID3Tag(lba,psize,(BYTE far *)&buf);
       dispfileinfo(TRUE,msg,playptr->filename,&CURRENTINFO);
       disptag(isTagOK,msg,&buf);
       /* ---------- */
#ifndef __BORLANDC__
       if (isTagOK && !isriff) size-=128;
       if (!PlayFile(lba,ofs,size)) break;
#else
       ISPLAYING=TRUE;
       disptime(FALSE,MP3INFOTIME_X,MP3INFOTIME_Y,sec);
#endif
     }
     while (ISPLAYING) {
       c=readkeypressed();
       switch (c) {
#ifdef __BORLANDC__
         case _KEY_OF_ALL   :
           ISPLAYING = FALSE;
           break;
#endif
         case _KEY_OF_VOLUP   :
 	       stVolTimeout=_GLOBAL_COUNTER_;
		   if (!isVolumeChanging) openvolume();
		   isVolumeChanging=TRUE;
           _setvolume(MAINVOLUME+1);
           setvolume(MAINVOLUME+1);
		   break;
         case _KEY_OF_VOLDOWN :
	       stVolTimeout=_GLOBAL_COUNTER_;
		   if (!isVolumeChanging) openvolume();
		   isVolumeChanging=TRUE;
           _setvolume(MAINVOLUME-1);
           setvolume(MAINVOLUME-1);
           break;
         case  _KEY_OF_EXEC :
           flashbtn(b1x,b1y,"tf");
           StopFile();
           playcnt=1; /**/
           break;
         case _KEY_OF_FUNC :
           _ENABLE_TIMER=FALSE;
           if (CURRENTDISCTYPE==_MIXED_DISC) {
             if (messagedlg("CDDAĐɐ؂ւ܂?",_MSGDLG_2BTN)) {
               StopFile();
               RESULT=_REQUEST_AUDIOMODE;
               ISPLAYING=FALSE;
             }
           } else
             messagedlg("I[fBIgbN͂܂.",_MSGDLG_NOBTN);
           _ENABLE_TIMER=TRUE;
           break;
         case -1 : /* NO KEY */
           if (isVolumeChanging) {
             cnt=((_GLOBAL_COUNTER_-stVolTimeout) & 0x7FFFFFFFUL) / 100;
             if (cnt>2) {
               closewindow();
		       isVolumeChanging=FALSE;
		     }
           }
           break;
         default:
           if (isVolumeChanging) closewindow();
		   isVolumeChanging=FALSE;
            break;
       } /* switch */
     } /* play wait while */
     if (isVolumeChanging) closewindow();
     if (ISLOADERROR) {
       RESULT=_REQUEST_CHANGEDISC;
       break;
     }
     if (RESULT) break; /**/
     index++;
     playcnt--;
     if (!playcnt) break;
     _ENABLE_TIMER=FALSE;
     DELAY10MS(200); /* WAIT 2sec */
     _ENABLE_TIMER=TRUE;
   } /* play list while */
   closewindow();
   if (MAINVOLUME!=PREVVOLUME) SaveVolume(MAINVOLUME);
   waitnotkeypressed();
   return(RESULT);
}

void PlayCDTr(int track)
{
   ISMP3=FALSE;
   if (track>TrackNum) return;
   SetDACMute(TRUE,FALSE);
   PlayCD(TrackLba[track],TrackLba[0]-TrackLba[track]);
}


#define CDDAINFO_X1 200
#define CDDAINFO_Y1 50
#define CDDAINFO_X2 400
#define CDDAINFO_Y2 150

#define CDDAINFOTIME_X 220
#define CDDAINFOTIME_Y 73

int cddaplay(int track)
{
   char msg[40];
   BOOL isVolumeChanging;
   DWORD stVolTimeout;
   int cnt,c,b1x,b1y;
   int i;
   BOOL ISPLAYBREAK;
   int RESULT;
   int PREVVOLUME;

   PREVVOLUME=MAINVOLUME;
   ISPLAYING=FALSE; RESULT=0;
   openwindow(CDDAINFO_X1,CDDAINFO_Y1,CDDAINFO_X2,CDDAINFO_Y2);
   draw_window_s(CDDAINFO_X1,CDDAINFO_Y1,CDDAINFO_X2,CDDAINFO_Y2);

   b1x=CDDAINFO_X1+_MSG_XOFS;   b1y=CDDAINFO_Y2-BTN_HEIGHT-_MSG_YOFS;
   drawbtn(b1x,b1y,"tf");

   PREVSEC=-65536;
   ISPLAYBREAK=FALSE;
   while (!ISPLAYBREAK) {
     sprintf(msg,"ygbN %dz",track);
     putmessage(CDDAINFO_X1+4,CDDAINFO_Y1+_MSG_YOFS,msg);
     disptime(TRUE,CDDAINFOTIME_X,CDDAINFOTIME_Y,0);

     PlayCDTr(track);
	 PLAYINGTRACK=track;
	 isVolumeChanging=FALSE; ISPLAYING=TRUE;

     while (ISPLAYING) {
       if (PLAYINGTRACK!=track) {
         track=PLAYINGTRACK;
         sprintf(msg,"ygbN %dz",track);
         putmessage(CDDAINFO_X1+4,CDDAINFO_Y1+_MSG_YOFS,msg);
       }
       c=readkeypressed();
       switch (c) {
         case _KEY_OF_VOLUP   :
 	       stVolTimeout=_GLOBAL_COUNTER_;
		   if (!isVolumeChanging) openvolume();
		   isVolumeChanging=TRUE;
           _setvolume(MAINVOLUME+1);
           setvolume(MAINVOLUME+1);
		   break;
         case _KEY_OF_VOLDOWN :
	       stVolTimeout=_GLOBAL_COUNTER_;
		   if (!isVolumeChanging) openvolume();
		   isVolumeChanging=TRUE;
           _setvolume(MAINVOLUME-1);
           setvolume(MAINVOLUME-1);
           break;
         case  _KEY_OF_EXEC :
           flashbtn(b1x,b1y,"tf");
           _ENABLE_TIMER=FALSE;
           PlayStop();
           _ENABLE_TIMER=TRUE;
           ISPLAYBREAK=TRUE;
           ISPLAYING=FALSE;
           break;
         case _KEY_OF_FUNC :
           _ENABLE_TIMER=FALSE;
           if (CURRENTDISCTYPE==_MIXED_DISC) {
             if (messagedlg("MP3Đɐ؂ւ܂?",_MSGDLG_2BTN)) {
               PlayStop();
               RESULT=_REQUEST_DATAMODE;
               ISPLAYING=FALSE;
               ISPLAYBREAK=TRUE;
             }
           } else
             messagedlg("f[^gbN͂܂.",_MSGDLG_NOBTN);
           _ENABLE_TIMER=TRUE;
           break;
         case _KEY_OF_STOP :
            PlayStop();
            ISPLAYBREAK=TRUE;
            ISPLAYING=FALSE;
            break;
         case -1 : /* NO KEY */
           if (isVolumeChanging) {
             cnt=((_GLOBAL_COUNTER_-stVolTimeout) & 0x7FFFFFFFUL) / 100;
             if (cnt>2) {
               closewindow();
		       isVolumeChanging=FALSE;
		     }
           }
           break;
         default:
           if (isVolumeChanging) closewindow();
		   isVolumeChanging=FALSE;
            break;
       } /* switch */
     } /* play wait while */
     if (isVolumeChanging) closewindow();
     if (!ISPLAYBREAK) {
       track++;
       if (track>TrackNum) {
         /* track=1; */
         break;
       }
       for (i=track;i<TrackNum;i++)
         if (isAudioTrack[i]) {
           track=i;
           break;
         }
     }
     if (!isDeviceOK()) {
       RESULT=_REQUEST_CHANGEDISC;
       break;
     }
   }
   closewindow();
   if (MAINVOLUME!=PREVVOLUME) SaveVolume(MAINVOLUME);
   SetDACMute(FALSE,FALSE);
   return(RESULT);
}

/* ------------------------------------------------------------------------- */

/* Set transfer address & size to DMAU registers. */
void SetDMAU(BYTE ch,void far *addr,DWORD size,BOOL IsInit)
{
  DWORD seg,ofs;
  BYTE mode;
  
  seg=FP_SEG(addr);  ofs=FP_OFF(addr);
  ofs=ofs+(seg << 4);
  size--; /*!!!*/
  if (IsInit) mode=0; /*Base & Current*/
  else        mode=4; /*Base only*/
  outp(_DMAU_DCH,mode | (ch & 0x3));
  /*Transfer Address*/
  outpw(_DMAU_DBAL,ofs & 0xffff);
  outp(_DMAU_DBAH,(ofs >> 16) & 0x0f);
  /*Transfer count*/
  outpw(_DMAU_DBCL,size);
}

/* Set transfermode of DMAU (loop or single transfer) */
void SetDMAUMode(BYTE ch,BOOL flg)
{
  outp(_DMAU_DCH,ch & 0x3);
  if (flg)
     outp(_DMAU_DMD,_DMAU71_MODE | _DMAU71_AUTI); /* Demand,Inc,Autoinit Enable,Mem->IO,Byte */
  else
     outp(_DMAU_DMD,_DMAU71_MODE); /* Demand,Inc,Autoinit Disable,Mem->IO,Byte */
}

/* ALL RESET DMAU */
void InitDMAU(void)
{
/*ALL*/
  outp(_DMAU_DICM,0x03); /* ALL Reset */
  outp(_DMAU_DMK,0x0f); /* ALL Channel Mask */
/*CH0(I2S DMA TRANSFER SETTING*/
  SetDMAUMode(_I2SDMA_CH,FALSE);
/*CH1 LCD */
  SetDMAUMode(_LCDDMA_CH,TRUE); /* CH1 Select */
  SetDMAU(_LCDDMA_CH,_VRAM_BASE,_VRAM_SIZE,TRUE);
/* ALL*/
  outp(_DMAU_DDCL,0x00); /*Normal Write mode,Enable DMA*/
  outp(_DMAU_DDCH,0x00); /*BUS Release Mode*/
  outp(_DMAU_DMK,inp(_DMAU_DMK) & ~_LCDDMA_MASK); /* LCDDMA Enable */
}

void ResetLCD(void)
{
  disp_onoff(FALSE); /* DISPLAY off */
  DELAY10MS(5);  /* wait 50ms */
  lcd_onoff(FALSE);
  DELAY10MS(15); /* wait 150ms */
/*STEAL DMA RESET*/
/*CH1 LCD */
  SetDMAUMode(_LCDDMA_CH,TRUE); /* CH1 Select */
  SetDMAU(_LCDDMA_CH,_VRAM_BASE,_VRAM_SIZE,TRUE);
/**/
  lcd_onoff(TRUE);
  DELAY10MS(15); /* wait 150ms */
  disp_onoff(TRUE); /* DISPLAY off */
}

/* software delay */
void v53_delay(int cnt)
{
   int i,j;
   for (j=0;j<cnt;j++) {
     for (i=0;i<1600;i++) {
      inp(FPGAR2);	// 160ns@12MHz
     }
   }
}

/* cnt ms wait routine */
void DELAY(int cnt)
{
#ifdef __BORLANDC__
  delay(cnt);
#else
  v53_delay(cnt);
#endif
}

/* cnt 10ms wait routine. this wait is worked by timer intr. */
void DELAY10MS(int cnt)
{
#ifdef __BORLANDC__
  delay(cnt*10);
#else
  DWORD st,diff;
  
  cnt++;
  st=_GLOBAL_COUNTER_;
  while (1) {
    diff=(_GLOBAL_COUNTER_-st) & 0x7fffffffUL;
    if (diff>=cnt) break;
  }
#endif
}


BOOL GetTrackInfo(BOOL *isExistDataTrack)
{
   BYTE far *buf;
   ATAPIPAC pac;
   int i,tr;
   BOOL isAudio,isData;;
   
   if (isExistDataTrack==NULL) return(FALSE);
   TrackNum=0; *isExistDataTrack=FALSE;
   buf=LocalBuf; isAudio=isData=FALSE;
   if (!isDeviceOK()) return(FALSE);
   clratapipac(&pac);
   pac.cmd=0x43;
   pac.param[5]=0x01;
   pac.param[6]=0x10;
   pac.param[7]=0x00;
#ifdef __BORLANDC__
   if (!SendATAPICommand(_DATAREAD,&pac,12,buf,0x1000)) {
     RequestSense(NULL);
     return(FALSE);
   }
#else
   if (!SendATAPICommand_INTR(&pac,12,buf,0x1000)) {
     RequestSense(NULL);
     return(FALSE);
   }
#endif
   TrackNum=buf[3];
   for (i=tr=0;i<=TrackNum;i++) {
     if (tr==TrackNum) tr=0;
     else			   tr=i+1;
     TrackLba[tr]=(DWORD)buf[4+i*8+7]+((DWORD)buf[4+i*8+6]<<8)
                +((DWORD)buf[4+i*8+5]<<16)+((DWORD)buf[4+i*8+4]<<24);
     if (tr) {
       if (buf[4+i*8+1] & 0x04) {
         isAudioTrack[tr]=FALSE;
     /*    RootLBA=TrackLba[tr]; */
         *isExistDataTrack=TRUE; /* track1 is data-track? */
         isData=TRUE;
       } else {
         isAudio=TRUE;
         isAudioTrack[tr]=TRUE;
       }
     }
   }
   if (isAudio && isData) CURRENTDISCTYPE=_MIXED_DISC;
   else if (isAudio) CURRENTDISCTYPE=_AUDIO_DISC;
   else CURRENTDISCTYPE=_DATA_DISC;
   return(TRUE);
}

void fstrcpy(char far *d,char far *s)
{
   while (*s) *d++=*s++;
   *d=0;
}

void CopyVolumeLabelUnicode(BYTE far *p)
{
  WORD code;
  int i;
  WORD far *wp;
  BYTE far *s;
  BYTE far *sp;
  
  s=sp=(BYTE far *)VOLUMELABELJ;
  wp=(WORD far *)p;
  for (i=0;i<16;i++) {
    code=*wp++;/* code is big-endian!! */
    if (!code) break;
    code=(code >> 8) | (code << 8);
    code=Unicode2SJIS(code);
    if ((code & 0xff00)==0) {
      *s++=code & 0xff;
      if ((code & 0xff)!=0x20) sp=s;
    } else if (code!=0xffff) {
      *s++=code >> 8;
      *s++=code & 0xff;
      sp=s;
    }
  }
  *sp=0;
}

void CopyVolumeLabel(BYTE far *p)
{
  WORD c;
  int i;
  BYTE far *s;
  BYTE far *sp;
  int isk1;
  
  s=sp=(BYTE far *)VOLUMELABEL;
  isk1=0;
  for (i=0;i<32;i++) {
    if (!*p) break;
    c=*p;
    if (isk1==1) {
      isk1++;
      sp=s;
      sp++;
    } else
    if (((c>=0x81) && (c<=0x9f))
        || ((c>=0xe0) && (c<=0xfc))) isk1=1;
    *s++=*p++;
    if ((c!=0x20) && !isk1) sp=s;
    if (isk1==2) isk1=0;
  }
  *sp=0;
}

int convchr(int n,BYTE far **p)
{
   int i,r;
   
   for (i=0,r=0;i<n;i++) {
     if ((**p>='0') && (**p<='9')) r=r*10+(**p-'0');
     else r=0;
     (*p)++;
   }
   return(r);
}

void SetVolumeTime(BYTE far *p)
{
   int yy,mm,dd;
   int hhh,mmm,sss,ms;

   yy=mm=dd=hhh=mmm=sss=0;
   yy=convchr(4,&p);
   mm=convchr(2,&p);
   dd=convchr(2,&p);
   hhh=convchr(2,&p);
   mmm=convchr(2,&p);
   sss=convchr(2,&p);
   ms=convchr(2,&p);
   sprintf(VOLUMEDATE,"%dN%d%d %d%d%db%d",yy,mm,dd,hhh,mmm,sss,ms);
}

WORD ConvSector(DWORD len)
{
   if (len==0) return(0);
   else return((len-1) >> 11)+1;
}

BOOL ReadPathTable(void)
{
   DWORD pathlen,pathlba;
   WORD pathsect;
   BYTE typ;
   int j,ii,jj;
   long i;
   BYTE far *pp;

   ISJOLIET=FALSE;
   /* analizing PVD,SVD. RootLBA is the offset of first track in last session */
   if (!GetLastSession(&RootLBA)) {
     RootLBA=0;
     return(FALSE);
   }
   for (i=0x10+RootLBA;;i++) {
     if (!LoadingData(LocalBuf,LOCAL_BUFSIZE,(DWORD)i,2048)) return(FALSE); /* Read PVD */
     DELAY(1);
     pp=LocalBuf; pp++;
     for (j=0;j<sizeof(ISO9660_ID);j++) if (ISO9660_ID[j]!=*pp++) break;
     if (j!=sizeof(ISO9660_ID)) return(FALSE);
     typ=*(BYTE far *)LocalBuf;
     if (typ==0xff) break; /* fin FF'CD001' */
     if ((typ==0x01) || (typ==0x02)) {
       RootDir=*(DWORD far *)(LocalBuf+156+2);
       RootDirLen=*(DWORD far *)(LocalBuf+156+10);
       pathlen=*(DWORD far *)(LocalBuf+132);
       pathlba=*(DWORD far *)(LocalBuf+140);
       if (typ==1) SetVolumeTime(LocalBuf+813);
       if (typ==0x02) {
         /* CHECK UCS-2 Escape sequence */
         for (ii=0;ii<3;ii++) {
           pp=LocalBuf+88;
           for (jj=0;jj<3;jj++) 
             if (JOLIETESC[ii][jj]!=*pp++) break;
           if (jj==3) break; 
         }
         if (jj==3) {
		   CopyVolumeLabelUnicode(LocalBuf+40);
           ISJOLIET=TRUE;
         }
         break;
       } else
		 CopyVolumeLabel(LocalBuf+40);
     }
   }
   if ((pathlen==0) || (pathlba<0x10)) return(FALSE);
   pathsect=ConvSector(pathlen);  /* conv byteorder-Length to sector */
   if (!ReadSector(pathlba,PathTbl,pathsect)) return(FALSE);
   return(TRUE);
}

/* copy filename without revision code */
void CopyFilename(BYTE far *s,BYTE far *p)
{
  WORD cnt;
  int i;
  int isk1;
  
  cnt=*p++; isk1=0;
  for (i=0;i<cnt;i++) {
    if (isk1==1) isk1++;
    else
    if (((*p>=0x81) && (*p<=0x9f))
        || ((*p>=0xe0) && (*p<=0xfc))) isk1=1;
    if (!isk1 && (*p==';')) break;
    *s++=*p++;
    if (isk1==2) isk1=0;
  }
  *s=0;
}

/* copy unicode filename without revision code */
/* and convert unicode to sjis-code            */
void CopyFilenameUnicode(BYTE far *s,BYTE far *p)
{
  WORD cnt,code;
  int i,len;
  WORD far *wp;
  
  cnt=*p++; len=0;
  if (cnt==1) *s++=*p;
  else {
    wp=(WORD far *)p;
    for (i=0;i<cnt >> 1;i++) {
      code=*wp++;/* code is big-endian!! */
      code=(code >> 8) | (code << 8);
      if (code==';') break;
      code=Unicode2SJIS(code);
      if ((code & 0xff00)==0) {
        *s++=code & 0xff;
        len++;
      } else if (code!=0xffff) {
        *s++=code >> 8;
        *s++=code & 0xff;
        len+=2;
      }
      if (len>63) break;
    }
  }
  *s=0;
}

const char mp3ext[3][4]={
  {'.','M','P','3'},{'.','W','A','V'},{'.','R','M','P'}} ;

/* Check '.mp3','.wav','.rmp' */
BOOL CheckMP3(BYTE far *s)
{
   int ckcnt[3];
   BYTE c;
   int isk1,j;
   
   isk1=0; 
   for (j=0;j<3;j++) ckcnt[j]=0;
   while (*s) {
     if (isk1==1) {
       isk1++;
       c=0;
     } else
       c=*s;
     if (((c>=0x81) && (c<=0x9f))
         || ((c>=0xe0) && (c<=0xfc))) isk1=1;
     if (isalpha(c)) c=toupper(c);
     for (j=0;j<3;j++) {
       if (!isk1 && (mp3ext[j][ckcnt[j]]==c)) ckcnt[j]++;
       else ckcnt[j]=0;
     }
     s++;
     if (isk1==2) isk1=0;
   }
   for (j=0;j<3;j++)
     if (ckcnt[j]==4) return(TRUE);
   return(FALSE);
}

/* Loading directory informations into global-directory-buffer */
BOOL LoadDirectory(DWORD lba,DWORD length,BOOL isJoliet)
{
   BYTE far *p;
   LPFILETYPE ptr;
   WORD cnt,next;
   
   DirFiles=0;
   if (!length) return(FALSE);
   if (!isDeviceOK()) return(FALSE);
   if (!LoadingData(LocalBuf,LOCAL_BUFSIZE,lba,length)) return(FALSE);
   p=LocalBuf; ptr=(LPFILETYPE)DirTbl;
   cnt=0;
   while (cnt<length) {
     next=*p;
     if (next==0) break;
     ptr->lba=*(DWORD far *)(p+2);   /*Extent pos*/
     ptr->size=*(DWORD far *)(p+10); /*Data Length*/
     ptr->flag=*(p+25); /* file type */
     /* filenames are stored by sjis-code (unicode is converted to sjis code) */
     if (!isJoliet)  CopyFilename((BYTE far *)(ptr->filename),p+32);
     else            CopyFilenameUnicode((BYTE far *)(ptr->filename),p+32);
     ptr->ismp3=CheckMP3(ptr->filename);
     p+=next;
     ptr=(LPFILETYPE)((BYTE far *)ptr+(DWORD)sizeof(FILETYPE));
     cnt+=next;
     DirFiles++;
     if (DirFiles==MAX_FILEDIR) break; /* max file check. */
   }
   return(TRUE);
}

/* make folder/file item string */
void MakeDispStr(int n,char *s,LPFILETYPE p)
{
   char typebuf[10];
   char filebuf[80];
   int siz;

   if (p->flag & 2) strcpy(typebuf," <DIR> ");
   else strcpy(typebuf,"       ");
   fstrcpy(filebuf,p->filename);
   if (filebuf[0]==1) strcpy(filebuf,"<ParentDir>");
   else if (filebuf[0]==0) strcpy(filebuf,"<CurrentDir>");
   siz=p->size >> 10;
   if (!siz) siz=1;
   if (strlen(filebuf)>20) {
     if (p->flag & 2) {
       CutStr(20,filebuf);
       sprintf(s,"%.2d. %-20s %s %dK",n,filebuf,typebuf,siz);
     } else {
       CutStr(28,filebuf);
       sprintf(s,"%.2d. %-28s %dK",n,filebuf,siz);
     }
   } else
     sprintf(s,"%.2d. %-20s %s %dK",n,filebuf,typebuf,siz);
}

/* Directory Listing */
BOOL DispDirectory2(int top)
{
   LPFILETYPE ptr,pptr;
   int i,col,items;
   int index;

   if (top>=DirFiles) return(TRUE);
/*   steal_onoff(FALSE);*/
   TrueDirFiles=0; items=0;
   /* CLEAR AREA */
   Grph_FillArea(1,1,SCR_WIDTH-3,MAX_DISPLINE * RAW_SIZE + 1,_WRITEMODE_RPL,0);
   pptr=ptr=(LPFILETYPE)DirTbl; /* start folder */
   for (i=0;i<MAX_DIRMAP;i++) DIRMAP[i]=NULL; /* clear all */
   for (i=0,col=0;i<DirFiles;i++) {
     if (*(ptr->filename)) {
       if ((pptr->lba==ptr->lba) && (*(ptr->filename)==1)) ;
       else {
	     if (col<MAX_DISPLINE*2) {
           if (items>=top) {
             DIRMAP[col]=ptr;
             MakeDispStr(items+1,dispbuf,ptr);
             if (col<MAX_DISPLINE)
               putmessage(DISP_COL1_S+1,col * RAW_SIZE + 2 ,dispbuf);
             else
               putmessage(DISP_COL2_S+1,(col - MAX_DISPLINE) * RAW_SIZE + 2 ,dispbuf);
             /* if this file in playlist,set half cursor */
             if ((ptr->flag & 2)==0) {
               index=getindex(ptr->lba,ptr->size);
               if (index>=0)
                 DispCursor(items-top,-1,MAX_DISPLINE,3);
             }
             col++;
           }
         }
         items++;
       }
     }
     ptr=(LPFILETYPE)((BYTE far *)ptr+(DWORD)sizeof(FILETYPE));
   }
   TrueDirFiles=items;
/*   steal_onoff(TRUE);*/
   return(TRUE);
}

/* Set All Files in current folder into playlist */
void SetAllDir(BOOL isClear)
{
   LPFILETYPE ptr;
   int i;
   int index;

   ptr=(LPFILETYPE)DirTbl; /* start folder */
   for (i=0;i<DirFiles;i++) {
     if ((ptr->flag & 2)==0) {
       index=getindex(ptr->lba,ptr->size);
       if (!isClear) {
         if (index<0)
           addplaylist(ptr->lba,ptr->size,ptr->filename);
       } else {
         if (index>-1)
           deleteplaylist(index);
       }
     }
     ptr=(LPFILETYPE)((BYTE far *)ptr+(DWORD)sizeof(FILETYPE));
   }
}

void MakeDispTrStr(int tr,long trlen,char *s)
{
  long mm,ss;
  
   mm=trlen / (75*60);
   ss=(trlen - (mm * 75 * 60)) / 75;
   if (isAudioTrack[tr])
      sprintf(s,"ygbN    %.2d   %.02d:%.02d",tr,(int)mm,(int)ss);
   else
      sprintf(s,"f[^gbN  %.2d   %.02d:%.02d",tr,(int)mm,(int)ss);
}

/* Track Listing */
BOOL DispTrack(int top)
{
   int i,col,items,tr;
   long trlen;

   if (top>=TrackNum) return(TRUE);
/*   steal_onoff(FALSE);*/
   items=0;
   /* CLEAR AREA */
   Grph_FillArea(1,1,SCR_WIDTH-3,MAX_DISPLINE_AUDIO * RAW_SIZE + 1,_WRITEMODE_RPL,0);
   for (i=0;i<MAX_DIRMAP;i++) TRMAP[i]=NULL; /* clear all */
   for (i=0,col=0;i<TrackNum;i++) {
     tr=items+1;
	 if (col<MAX_DISPLINE_AUDIO*2) {
       if (items>=top) {
         TRMAP[col]=tr;
         if (items==TrackNum-1) trlen=TrackLba[0]-TrackLba[tr];
         else	                trlen=TrackLba[tr+1]-TrackLba[tr];
         MakeDispTrStr(tr,trlen,dispbuf);
         if (col<MAX_DISPLINE_AUDIO)
           putmessage(DISP_COL1_S+1,col * RAW_SIZE + 2 ,dispbuf);
         else
           putmessage(DISP_COL2_S+1,(col - MAX_DISPLINE_AUDIO) * RAW_SIZE + 2 ,dispbuf);
         col++;
       }
     }
     items++;
   }
/*   steal_onoff(TRUE);*/
   return(TRUE);
}

void GetCsr(int csr,int max,int *x1,int *y1,int *x2,int *y2)
{
   if (csr<max) {
     *x1=DISP_COL1_S; *x2=DISP_COL1_E;
   } else {
     *x1=DISP_COL2_S; *x2=DISP_COL2_E;
   }
   if (csr>=max) csr-=max;
   *y1=csr * RAW_SIZE + 2;  *y2=*y1+16;
}

void DispCursor(int prevcsr,int csr,int max,int depth)
{
   int x1,x2,y1,y2;
   
   if (prevcsr>-1) {
     GetCsr(prevcsr,max,&x1,&y1,&x2,&y2);
     Grph_FillArea(x1,y1,x2,y2,_WRITEMODE_XOR,depth);
   }
   if (csr>-1) {
     GetCsr(csr,max,&x1,&y1,&x2,&y2);
     Grph_FillArea(x1,y1,x2,y2,_WRITEMODE_XOR,depth);
   }
}

void InitTimer(void)
{
  /* TCU0 0,1 SQUARE WAVE MODE, ch2 = RateGen Mode, F=1/(24.576MHz / 4 ) * (Cnt+1) for ShiftClock */
  di();
  outp(_TIMER_CTRL,0x36); outp(_TIMER_CH0,0x80); outp(_TIMER_CH0,0x02); /* LCDCK Timer CH0(0.1ms) */
  outp(_TIMER_CTRL,0x56); outp(_TIMER_CH1,0x6); /* DSP-SERIAL CK Timer CH1(F=1us) */
  outp(_TIMER_CTRL,0xB4); outp(_TIMER_CH2,0x00); outp(_TIMER_CH2,0xF0); /* Timer CH2(10ms) */
  _TIMER_DIV=1; _ENABLE_TIMER=TRUE;  /*CH2*/
  outp(_ICU_MASK,inp(_ICU_MASK) & ~_TIMER_INTR_MASK); /* timer interrupt enable !! */
/*  outp(_TIMER_CTRL,0x56); outp(_TIMER_CH1,0x7);*/
  outp(_8255C,0x92); /* Set Mode0 PA,PB = input , PC = all output */
  ei();
}

long REMAINDATAS;
long LOAD_COUNT[2];
DWORD LBA;
BOOL ALREADY_ENTERED;
int PLAYEDBUFF;

#define LOADBUFSIZE (0x10000UL-2048)

/* /TC interrupt routine */
/* This routine works to read datas from cd-rom, and set next play address to DMA base register */
void dma_intr(void)
{
	long loadsect;
    DWORD restsize;
    BYTE far *p;

    /* this routine called from DMA_INTR_HD@start.p86 */
	/* EOI of IIC has been finished yet (@DMA_INTR_HD) */
    if (ALREADY_ENTERED) return;  /* aboid re-entrance */
    ALREADY_ENTERED=TRUE;  /* Flag on */
	while (inp(_DMAU_DST) & _I2SDMA_MASK) { /* Check SRLDMARQ TC(others:LCDDMARQ is ignored) */
	  if (REMAINDATAS==0) {
	    /*FINISH*/
        switch (PLAYINGSTAT) {
          case _PLAYING:
				         SetDMAUMode(_I2SDMA_CH,FALSE); /* Demand,Inc,Autoinit Disable,Mem->IO,Byte */
     					 outp(_DMAU_DMK,inp(_DMAU_DMK) & ~_I2SDMA_MASK);
           			     PLAYINGSTAT=_LAST_BUFFER_PLAYING;
					     break;
		  case _LAST_BUFFER_PLAYING :
					     outp(_DMAU_DMK,inp(_DMAU_DMK) | _I2SDMA_MASK);
                         outp(_ICU_MASK,inp(_ICU_MASK) | _DMAU_INTR_MASK);
                         LOAD_COUNT[0]=LOAD_COUNT[1]=0;
                         PLAYINGSTAT=_STOPPLAYING;
                         ISPLAYING=FALSE;
                         break;
        }
	  } else {
		/* */
        PLAYEDBUFF++; PLAYEDBUFF&=1;
		if (REMAINDATAS>=LOADBUFSIZE) LOAD_COUNT[PLAYEDBUFF]=LOADBUFSIZE;
		else LOAD_COUNT[PLAYEDBUFF]=REMAINDATAS;
		loadsect=((LOAD_COUNT[PLAYEDBUFF]-1) >> 11)+1;
		ei(); /*while loading data from CD-ROM ... */
		if (!LoadingData(loadbuff[PLAYEDBUFF],LOADBUFSIZE,LBA,loadsect*2048)) {
		  /*READ ERROR!! FORCE STOP PLAYING CURRENT FILE */
		  di();
		  StopFile();
		  ISLOADERROR=TRUE;
		  break;
		};
		di();
		LBA+=(DWORD)loadsect; REMAINDATAS-=LOAD_COUNT[PLAYEDBUFF];
        restsize=LOADBUFSIZE-LOAD_COUNT[PLAYEDBUFF];
        if (restsize>0) {
          if (LOAD_COUNT[PLAYEDBUFF]<0x100) {
            p=loadbuff[PLAYEDBUFF]+LOAD_COUNT[PLAYEDBUFF];
            far_memset(p,0,restsize);
            LOAD_COUNT[PLAYEDBUFF]=0x100;
          }
        }
        SetDMAU(_I2SDMA_CH,loadbuff[PLAYEDBUFF],LOAD_COUNT[PLAYEDBUFF],FALSE);
	  }
    }
    ALREADY_ENTERED=FALSE;
}

void StopFile(void)
{
#ifndef __BORLANDC__
   /* Set DMARQ Mask & /TC intr disable*/
   outp(_DMAU_DMK,inp(_DMAU_DMK) | _I2SDMA_MASK);
   outp(_ICU_MASK,inp(_ICU_MASK) | _DMAU_INTR_MASK);
   SetDACMute(FALSE,FALSE); /* MP3 & CDDA Mute */
   DELAY(10);
   outp(FPGAR1,inp(FPGAR1) & ~(DSPEN | SDMASK));
#endif
   LOAD_COUNT[0]=LOAD_COUNT[1]=0;
   PLAYINGSTAT=_STOPPLAYING; ISPLAYING=FALSE;
}

BOOL PlayFile(DWORD lba,DWORD ofs,DWORD size)
{
   long loadsect;
   int i;

   /* Set DMARQ Mask & /TC intr disable*/
   SetDACMute(FALSE,FALSE); /* MP3 & CDDA Mute */
   outp(_DMAU_DMK,inp(_DMAU_DMK) | _I2SDMA_MASK);
   outp(_ICU_MASK,inp(_ICU_MASK) | _DMAU_INTR_MASK);
   DELAY10MS(10); /* WAIT 100 MS */
   outp(FPGAR1,inp(FPGAR1) | DSPEN);  /* WSEN ENABLE */

/*   |<----------- true read size ------------>|  */
/*                |<-------- play size ------->|  */
/*   |<--- ofs -->|                               */
   REMAINDATAS=size+ofs;  LBA=lba; PLAYEDBUFF=1; /* init global info */
   LOAD_COUNT[0]=LOAD_COUNT[1]=0;  PLAYINGSTAT=_PLAYING; ISLOADERROR=FALSE;
   for (i=0;i<2;i++) {
     if (REMAINDATAS>0) {
  	   if (REMAINDATAS>=LOADBUFSIZE) LOAD_COUNT[i]=LOADBUFSIZE;
	   else LOAD_COUNT[i]=REMAINDATAS;
	   loadsect=((LOAD_COUNT[i]-1) >> 11)+1;
       if (!LoadingData(loadbuff[i],LOADBUFSIZE,LBA,loadsect*2048)) {
         ISLOADERROR=TRUE;
         StopFile();
         return(FALSE);
       } else
         DELAY(2);
	   LBA+=(DWORD)loadsect; REMAINDATAS-=LOAD_COUNT[i];
     }
   }
   _MP3_STARTTIME=_GLOBAL_COUNTER_; ISPLAYING=TRUE; ISMP3=TRUE;
   if (LOAD_COUNT[1]==0) {
     /*start(loaddata<=LOADBUFSIZE)*/
     SetDMAUMode(_I2SDMA_CH,FALSE); /* Demand,Inc,Autoinit Disable,Mem->IO,Byte */
     SetDMAU(_I2SDMA_CH,loadbuff[0]+ofs,LOAD_COUNT[0]-ofs,TRUE);
     PLAYINGSTAT=_LAST_BUFFER_PLAYING; /*FIN*/
     outp(_DMAU_DMK,inp(_DMAU_DMK) & ~_I2SDMA_MASK);
     outp(_ICU_MASK,inp(_ICU_MASK) & ~_DMAU_INTR_MASK);
     outp(FPGAR1,inp(FPGAR1) | SDMASK);  /* ENABLE */
     DELAY10MS(7); /* WAIT 70 MS */
     SetDACMute(FALSE,TRUE); /* MP3 Enable, CDDA Mute */
   } else {
     /*start(LOADBUFSIZE)*/
     SetDMAUMode(_I2SDMA_CH,TRUE); /* Demand,Inc,Autoinit Enable,Mem->IO,Byte */
     SetDMAU(_I2SDMA_CH,loadbuff[0]+ofs,LOAD_COUNT[0]-ofs,TRUE);
     outp(_DMAU_DMK,inp(_DMAU_DMK) & ~_I2SDMA_MASK);
     outp(_ICU_MASK,inp(_ICU_MASK) & ~_DMAU_INTR_MASK);
     outp(FPGAR1,inp(FPGAR1) | SDMASK);  /* ENABLE */
     DELAY10MS(7); /* WAIT 70 MS */
     SetDACMute(FALSE,TRUE); /* MP3 Enable, CDDA Mute */
     SetDMAU(_I2SDMA_CH,loadbuff[1],LOAD_COUNT[1],FALSE); /* NEXT BLOCK */
   }
   return(TRUE);
}

/* 10ms timer hardware interrupt */
void timer_intr(void)
{
   DWORD diffsec;
   int tr;
   long sec;

   _GLOBAL_COUNTER_++;
   _TIMER_DIV++;
   if (_TIMER_DIV>40) _TIMER_DIV=1;
   if (ISPLAYING) {
     if (ISMP3) {
       /* MP3 PLAYER */
       /* this code executes every 0.2 sec */
       if (_TIMER_DIV % 20)  return;
       if (!_ENABLE_TIMER) return; /* if dont use I2C bus */
       diffsec=(_GLOBAL_COUNTER_ - _MP3_STARTTIME) & 0x7FFFFFFF;
       diffsec=diffsec / 100;
       disptime(FALSE,MP3INFOTIME_X,MP3INFOTIME_Y,diffsec);
     } else {
      /* CD PLAYER */
       /* this code executes every 0.4 sec */
       if (_TIMER_DIV % 40)  return;
       if (!_ENABLE_TIMER) return; /* if dont use adapi */
       tr=isPlaying(&sec);
       if (tr<0) ISPLAYING=FALSE;
       else {
         PLAYINGTRACK=tr;
         sec=sec / 75;
         disptime(TRUE,CDDAINFOTIME_X,CDDAINFOTIME_Y,sec);
       }
     }
   }
}


void DispAudioDiscInfo(void)
{
	char msg[80];
	long mm,ss;
	
	sprintf(msg,"ySgbNz %d gbN",TrackNum);
	putmessage(10,200-RAW_SIZE-5,msg);

    mm=TrackLba[0] / (75*60);
    ss=(TrackLba[0] - 75*60*mm) / 75;
	sprintf(msg,"yStԁz %.02d %.02db",(int)mm,(int)ss);
	putmessage(320,200-RAW_SIZE-5,msg);
}

void DispVolumeInfo(BOOL isJoliet)
{
	char msg[80];
	
	strcpy(msg,"y{[xz ");
    if (isJoliet) strcat(msg,VOLUMELABELJ);
    else          strcat(msg,VOLUMELABEL);
	putmessage(10,200-RAW_SIZE*2-6,msg);

	strcpy(msg,"yt@CVXe^Cvz ");
	if (isJoliet) strcat(msg,"Joliet");
	else          strcat(msg,"ISO9660");
	putmessage(10,200-RAW_SIZE-5,msg);
	
	strcpy(msg,"y쐬z ");
    strcat(msg,VOLUMEDATE);
	putmessage(300,200-RAW_SIZE-5,msg);
}

/* Cursor key control(generic) */
void CursorKey(int c,int *csr,int *top,int filemaxptr,int max)
{
   switch (c) {
       case _KEY_OF_UP    : 
                            if (*csr>1) {
                              (*csr)--;
                            } else {
                              if (*csr+*top>0) {
                                *top-=max;
                                *csr=max-1;
                                if (*csr+*top<0) *csr=*top=0;
                              }
       						}
                            break;
       case _KEY_OF_DOWN  : if (*csr<max*2-1) {
                              if (filemaxptr>*csr+*top) (*csr)++; /**/
       						} else {
                              if (filemaxptr>*csr+*top) {
                                *top+=max; /**/
                                *csr=max; /*next page *top*/
                              }
                              if (filemaxptr<*csr+*top) *csr=filemaxptr-*top; /**/
       						}
                            break;
       case _KEY_OF_RIGHT : if (*csr<max) {
                              *csr+=max;
                              if (filemaxptr<*csr+*top) *csr=filemaxptr-*top;
                            } else {
                              if (filemaxptr>=*top+max*2) {
                                *top+=max;
                                if (*csr+*top>filemaxptr) *csr=filemaxptr-*top;
                              }
                            }
                            break;
       case _KEY_OF_LEFT  : 
                            if (*csr>=max) {
                              *csr-=max;
                            } else {
                              if (*top>0) *top-=max;
                              if (*top<0) *top=0;
                            }
                            break;
   }

}

/* THIS LOOP is MP3 MAIN PROGRAM LOOP */
int MP3_MENU(void)
{
   DWORD lba,length,prevlba;
   int c,top,prevtop,csr,prevcsr;
   LPFILETYPE ptr;
   int filemaxptr;
   int index,rslt;
   BOOL isSetAll;

   clearplaylist();
   if (!ReadPathTable()) return(_REQUEST_CHANGEDISC); /* recognize joliet,iso9660 ... */
   DELAY(1);
   DispVolumeInfo(ISJOLIET);
   csr=0; lba=RootDir; length=RootDirLen; prevcsr=-1;
   if (!LoadDirectory(lba,length,ISJOLIET)) return(_REQUEST_CHANGEDISC);
   prevlba=lba; top=-1; prevtop=0; isSetAll=FALSE;
   while (1) {
/* ------------- */
     /* Directory Change. if changed then must be refresh all disp.  */
     if (prevlba!=lba) {
       if (!LoadDirectory(lba,length,ISJOLIET)) return(_REQUEST_CHANGEDISC);
       else top=-1;
     }
     prevlba=lba;
     /* All Disp */
     if (prevtop!=top) {
       if (top<0) {
         top=0;
         DispCursor(prevcsr,-1,MAX_DISPLINE,10);
         csr=-1; prevcsr=-2;
       }
       DispDirectory2(top);
       DispCursor(prevcsr,-1,MAX_DISPLINE,10);
       /* if prevcsr<0 then previouse cursor not exist. */
     }
     prevtop=top;
     if (prevcsr!=csr) {
       if (csr<0) csr=0;
       DispCursor(prevcsr,csr,MAX_DISPLINE,10);
     }
     prevcsr=csr;
     filemaxptr=TrueDirFiles-1;
/* ------------- */
     c=readkey();
     if (c!=_KEY_OF_ALL) isSetAll=FALSE; /**/
     switch (c) {
       case _KEY_OF_UP    : 
       case _KEY_OF_DOWN  : 
       case _KEY_OF_RIGHT : 
       case _KEY_OF_LEFT  : 
						    CursorKey(c,&csr,&top,filemaxptr,MAX_DISPLINE);
                            break;
       case _KEY_OF_PLAY  : if (PLAYLISTCNT>0) {
                              if (messagedlg("vCXg̃t@C̉tJn܂H",_MSGDLG_2BTN)) {
                                if (!isDeviceOK()) return(_REQUEST_CHANGEDISC);
                                  rslt=mp3play(0,-1);
                                  if (rslt) return(rslt);
                              }
                            }
                            break;
       case _KEY_OF_EXEC  : if (DIRMAP[csr]) {
                              ptr=DIRMAP[csr];
                              if (ptr->flag & 2) lba=ptr->lba;
                              else {
                                if (ptr->ismp3) {
                                  if (messagedlg("̃t@C̉tJn܂H",_MSGDLG_2BTN)) {
                                    if (!isDeviceOK()) return(_REQUEST_CHANGEDISC);
                                    index=addplaylist(ptr->lba,ptr->size,ptr->filename);
                                    rslt=mp3play(index,1);
                                    deleteplaylist(index);
                                    if (rslt) return(rslt);
                                  }
                                }
                              }
       						}
                            break;
       case _KEY_OF_ALL    : 
                            if (isSetAll) SetAllDir(FALSE);
                            else          SetAllDir(TRUE);
                            isSetAll=!isSetAll;
                            prevtop=-1; /* all disp refresh */
                            break;
       case _KEY_OF_SELECT :
                            if (DIRMAP[csr]) {
                              ptr=DIRMAP[csr];
                              if ((ptr->flag & 2)==0) {
                                index=getindex(ptr->lba,ptr->size);
                                if (index>-1) {
                                  /* already exists*/
                                  deleteplaylist(index);
                                } else {
                                  addplaylist(ptr->lba,ptr->size,ptr->filename);
                                }
                                DispCursor(prevcsr,-1,MAX_DISPLINE,10);
                                DispCursor(prevcsr,-1,MAX_DISPLINE,3);
                                DispCursor(prevcsr,-1,MAX_DISPLINE,10);
                                CursorKey(_KEY_OF_DOWN,&csr,&top,filemaxptr,MAX_DISPLINE);
                              }
                            }
                            break;
       case _KEY_OF_FUNC  : if (CURRENTDISCTYPE==_MIXED_DISC) return(_REQUEST_AUDIOMODE);
                            break;
     }
   }
}

/* THIS LOOP is MAIN PROGRAM LOOP */
int CDDA_MENU(void)
{
   int c,top,prevtop,csr,prevcsr;
   int filemaxptr,rslt;

   csr=0; prevcsr=-1; top=-1; prevtop=0;
   DispAudioDiscInfo();
   while (1) {
/* ------------- */
     /* All Disp */
     if (prevtop!=top) {
       if (top<0) {
         top=0;
         DispCursor(prevcsr,-1,MAX_DISPLINE_AUDIO,10);
         csr=-1; prevcsr=-2;
       }
       DispTrack(top);
       DispCursor(prevcsr,-1,MAX_DISPLINE_AUDIO,10);
       /* if prevcsr<0 then previouse cursor not exist. */
     }
     prevtop=top;
     if (prevcsr!=csr) {
       if (csr<0) csr=0;
       DispCursor(prevcsr,csr,MAX_DISPLINE_AUDIO,10);
     }
     prevcsr=csr;
     filemaxptr=TrackNum-1;
/* ------------- */
     c=readkey();
     switch (c) {
       case _KEY_OF_UP    : 
       case _KEY_OF_DOWN  : 
       case _KEY_OF_RIGHT : 
       case _KEY_OF_LEFT  : 
						    CursorKey(c,&csr,&top,filemaxptr,MAX_DISPLINE_AUDIO);
                            break;
       case _KEY_OF_FUNC  : if (CURRENTDISCTYPE==_MIXED_DISC) return(_REQUEST_DATAMODE);
       case _KEY_OF_EXEC  : if (TRMAP[csr]) {
                              if (isAudioTrack[TRMAP[csr]]) {
                                rslt=cddaplay(TRMAP[csr]);
  							    if (rslt) return(rslt);
  							  } else {
  							    messagedlg("ygbNł͂܂.",_MSGDLG_1BTN);
  							  }
       						}
                            break;
#ifdef __BORLANDC__
       case _KEY_OF_ESC   : 
                            PlayStop();
                            return(TRUE);
#endif
     }
   }
}


/* THIS LOOP is MAIN PROGRAM LOOP */
void mainmenu(void)
{
   BOOL isDataDisc;
   BOOL reqAudio;
   int rslt;
   
   reqAudio=FALSE;
   while (1) {
     while (!GetTrackInfo(&isDataDisc)) {
       messagedlg("foCXłĂ܂B fBXN}ĂB",_MSGDLG_1BTN);
     }
     while (1) {
       if (isDataDisc && !reqAudio) {
         dispbackground();
         rslt=MP3_MENU();
       } else {
         /* this disc is audio only. */
         dispbackground_audio();
         rslt=CDDA_MENU();
       }
       if (rslt==_REQUEST_CHANGEDISC) break;
       else if (rslt==_REQUEST_AUDIOMODE) reqAudio=TRUE;
       else if (rslt==_REQUEST_DATAMODE) reqAudio=FALSE;
       else if (!isDeviceOK()) break;
	   else return; /* return! */
     }
   }
}

/* Initialize font & memories */
void initmemory(void)
{
   BYTE dat;

#ifdef __BORLANDC__
/*----*/
   FontFile=fopen("afont.fnt","rb");
   if (FontFile) {
     fseek(FontFile,17,SEEK_SET);
     fread(LocalAnkBuf,sizeof(LocalAnkBuf),1,FontFile);
     fclose(FontFile);
   }
   UnicodeFile=fopen("unicode.bin","rb");
   FontFile=fopen("kfont.fnt","rb");
   if (FontFile) {
     fseek(FontFile,17,SEEK_SET);
     fread(&dat,1,1,FontFile);
     fread(KanjiMap,(WORD)dat*4,1,FontFile);
     FontOfs=(WORD)dat*4+18;
   }
/*----*/
   PathTbl=(BYTE far *)farmalloc(0x8000);
   DirTbl=(BYTE far *)farmalloc(0x8000);
   LocalBuf=(BYTE far *)farmalloc(0xffe0);
   PlayList=(LPPLAYLIST)farmalloc(0x4000);
   WinStack[0].buf=(void far *)farmalloc(0x4000);
   WinStack[1].buf=(void far *)farmalloc(0x4000);
#else
/*----*/
   memset(KanjiMap,0,sizeof(KanjiMap));
   KanjiRom=(BYTE far *)MK_FP(0xA000,0);
   AnkRom=(BYTE far *)MK_FP(0xDF00,0);
   dat=*(BYTE far *)(0xA0000011UL);  /* Font Table Size */
   far_memcpy(KanjiMap,(BYTE far *)0xA0000012UL,dat*4); /*copy font map */
   FontOfs=dat*4+18;
/*----*/
   PathTbl=(BYTE far *)MK_FP(0x3000,0); /*32k*/
   DirTbl=(BYTE far *)MK_FP(0x3800,0); /*32k*/
   LocalBuf=(BYTE far *)MK_FP(0x4000,0); /*64k*/
   PlayList=(LPPLAYLIST)MK_FP(0x7C00,0); /*16k*/
   loadbuff[0]=(BYTE far *)MK_FP(0x5000,0); /*64k*/
   loadbuff[1]=(BYTE far *)MK_FP(0x6000,0); /*64k*/
   WinStack[0].buf=(void far *)MK_FP(0x7400,0); /*16k*/
   WinStack[1].buf=(void far *)MK_FP(0x7800,0); /*16k*/
#endif
}

/* for dosdebug. deallocate working memories & close font files.*/
void freememory(void)
{
#ifdef __BORLANDC__
   if (FontFile) fclose(FontFile);
   if (UnicodeFile) fclose(UnicodeFile);
   farfree(WinStack[0].buf);
   farfree(WinStack[1].buf);
   farfree(PathTbl);
   farfree(DirTbl);
   farfree(LocalBuf);
#endif
}

void initglobalvariables(void)
{
   ALREADY_ENTERED=FALSE; ENABLEDMA=FALSE; ISJOLIET=ISPLAYING=ISMP3=ISEMPH=FALSE;
   MAINVOLUME=LoadVolume(); /* load configration */
   PLAYLISTCNT=0; PLAYINGSTAT=_STOPPLAYING;
   PathTbl=NULL; RootDir=0; TrackNum=0; Loadedsize=0; RootLBA=0; isOpenNum=0;
   _GLOBAL_COUNTER_=_MP3_STARTTIME;
}

void dispwaitmessage(void)
{
  messagedlg("CD-ROMĂ܂EEE",_MSGDLG_NOBTN_NOCLOSE);
}

void closewaitmessage(void)
{
  closewindow();
}

void main(void)
{
   initglobalvariables();
   initmemory();
#ifndef __BORLANDC__
   Grph_Init();
   Grph_FillArea(0,0,SCR_WIDTH-1,SCR_HEIGHT-1,_WRITEMODE_RPL,5); /* chidori clear */
   dispwaitmessage();
   InitTimer(); DACDSPReset(); InitDMAU();
   _setvolume(MAINVOLUME);
   lcd_onoff(TRUE);   /* LCD STEAL ON */
   DELAY10MS(10);     /* 100ms WAIT */
   disp_onoff(TRUE);  /* LCD DISPLAY ON */
   outp(FPGAR1,inp(FPGAR1) & ~RESETCD);
   DELAY10MS(10);     /* 100ms CD-ROM RESET# Enable */
   outp(FPGAR1,inp(FPGAR1) | RESETCD);
   DELAY10MS(20);
   while (inpw(FPGAR2) & PDIAG); /* PDIAG CHECK */
   ATAReset();
   DELAY10MS(10);     /* 100ms Wait */
   RequestSense(NULL); /* Reset status clear */
   SetFeature(0x0c);
   DELAY10MS(600);    /* WAIT 6SEC */
   closewaitmessage(); /* close wait message */
   isDeviceOK(); DELAY10MS(10); isDeviceOK(); DELAY10MS(10); /*DUMMY CALL*/
   if (readkeypressed()==_KEY_OF_FUNC) monitor();
   else mainmenu();
#else
   monitor();
#endif
   freememory();
}
