/* MPROM Hardware definations (for FPGA & V53A inside logics) */
/* programmed by M.yanagisawa(MoroboshiLUM) */

#ifdef __BORLANDC__

#define ATA_DATA  0x170
#define ATA_FERT  0x171
#define ATA_INTR  0x172
#define ATA_SECC  0x172
#define ATA_SECN  0x173
#define ATA_CLCL  0x174
#define ATA_BYCL  0x174
#define ATA_CLCH  0x175
#define ATA_BYCH  0x175
#define ATA_DVHD  0x176
#define ATA_CMD   0x177
#define ATA_STAT  0x177

#define ATA_DVCL  0x376
#define ATA_ASTAT 0x376

#define ATA_OUTP(pn,dat) outp(pn,dat)
#define ATA_INP(pn)  inp(pn)

#else

#define ATA_DATA  0x100
#define ATA_FERT  0x102
#define ATA_INTR  0x104
#define ATA_SECC  0x104
#define ATA_SECN  0x106
#define ATA_CLCL  0x108
#define ATA_BYCL  0x108
#define ATA_CLCH  0x10A
#define ATA_BYCH  0x10A
#define ATA_DVHD  0x10C
#define ATA_CMD   0x10E
#define ATA_STAT  0x10E

#define ATA_DVCL  0x20C
#define ATA_ASTAT 0x20C

#define ATA_OUTP(pn,dat) outpw(pn,(dat) & 0xff)
#define ATA_INP(pn)  (inpw(pn) & 0xff)

#endif

/* 8255 */
#define _8255C   0x703
#define _8255D_A 0x700
#define _8255D_B 0x701
#define _8255D_C 0x702

/* FPGA Control Registers */
#define FPGAR1 0x300
#define FPGAR2 0x400
#define FPGAR3 0x500
#define FPGAR4 0x600

/* FPGAR1 */
#define I2CCOUT  0x80
#define I2CDOUT  0x40
#define DSPEN    0x20
#define RESETDSP 0x10
#define RESETCD  0x08
#define SDMASK   0x04
#define I2C_COM  0x02  /* FOR OUTPUT */
#define I2C_DATA 0x01  /* FOR OUTPUT */

/* FPGAR2 */
#define SREGEMP 0x80
#define DEMAND  0x04
#define INTRQ 0x08
#define PDIAG 0x04
/*#define I2C_COM  0x02  /* FOR INPUT */
/*#define I2C_DATA 0x01  /* FOR INPUT */

/* FPGAR3 */
/* I2S Serial Data Shift Register */

/* FPGAR4 */
#define LCDON   0x01    /* LCD LP/YD/XSCL Output Enable */
#define STEALON 0x02    /* LCD XSCL Output Enable(LCD enters stand-by mode) */
#define DOFF1   0x40	/* LCD Panel display on(left)  */
#define DOFF2   0x80	/* LCD Panel display on(right) */

/* V53A Inside DMAU(uPD71071 compatible) */
#define _DMAU_DICM 0xe000
#define _DMAU_DCH  0xe001
#define _DMAU_DBCL 0xe002
#define _DMAU_DBCH 0xe003
#define _DMAU_DBAL 0xe004
#define _DMAU_DBAM 0xe005
#define _DMAU_DBAH 0xe006
#define _DMAU_DDCL 0xe008
#define _DMAU_DDCH 0xe009
#define _DMAU_DMD  0xe00A
#define _DMAU_DST  0xe00B
#define _DMAU_DMK  0xe00F

/* V53A Inside DMAU(uPD71037 compatible) */
#define _DMAU_DRCA1 0xe002
#define _DMAU_DRCC1 0xe003
#define _DMAU_DRST  0xe008
#define _DMAU_DWC	0xe008
#define _DMAU_DWRQ  0xe009
#define _DMAU_DWSM  0xe00a
#define _DMAU_DWM   0xe00b
#define _DMAU_DCBP  0xe00c
#define _DMAU_DINT  0xe00d
#define _DMAU_DCM   0xe00e
#define _DMAU_DWAM  0xe00f

#define _DMAU_BANK  0xe040

#define _DMAU71_MODE 0x08  /* Demand,Inc,Autoinit Disable,Mem->IO,Byte */
#define _DMAU71_AUTI 0x10  /* AutoInitialize Flag */

#define _DMAU37_MODE 0x09  /* Demand,Inc,Autoinit Disable,Mem->IO,Byte */
#define _DMAU37_AUTI 0x10  /* AutoInitialize Flag */

#define _I2SDMA_CH 0
#define _I2SDMA_MASK (1 << _I2SDMA_CH)
#define _LCDDMA_CH 1
#define _LCDDMA_MASK (1 << _LCDDMA_CH)

/* V53A Inside PIC(uPD8259 compatible) */
#define _ICU_BASE  0xe010
#define _ICU_MASK  0xe011

#define _SIU_INTR_MASK 0x01
#define _DMAU_INTR_MASK 0x02
#define _CDROM_INTR_MASK 0x04
#define _TIMER_INTR_MASK 0x08

/* V53A Inside TIMER controler */
#define _TIMER_CH0 0xe020
#define _TIMER_CH1 0xe021
#define _TIMER_CH2 0xe022

#define _TIMER_CTRL 0xe023
