; ------------------------------------------------------------
; MP-ROM firmware BOOT & BIOS code
; Copyright (C) 2000 by M.yanagisawa(MoroboshiLUM)
; for TurboAssembler(Borland)
; ------------------------------------------------------------

	.186
	LOCALS
	JUMPS

INCLUDE	V53A.INC
INCLUDE	BIOS.INC

RXBUFSIZE       equ      2048    	; rs232c receive buffer (must be 2^x size)
FLOW_OFF_POINT  equ     (RXBUFSIZE / 4) *3
FLOW_ON_POINT   equ     (RXBUFSIZE / 4)

FLASH_PAGINGSEG	equ	0E000H

;; logical flah
_OFF	equ	0
_ON	equ	1

TRUE	equ	1
FALSE	equ	0

EXTADRVECT	equ	35H

EXTRN	C_INIT_:FAR			; LSIC-86 Startup Entry
EXTRN	_ENDSTACK:WORD			; LSIC-86 Stack bottom label
EXTRN UnicodeInt:FAR			; Unicode Conversion Bios
EXTRN DrawBiosInt:FAR			; Draw Bios
EXTRN GrphDataInt:FAR			; GraphicsData Bios


DGROUP	GROUP

XSTACK	SEGMENT WORD	STACK	'DATA'
XSTACK	ENDS

;;
;; Stacks and works for Assembler.
;;
SYSRAM	SEGMENT	WORD PUBLIC 'ASMDATA'
		ASSUME	DS:SYSRAM

		org	0

_EXECUTION_AREA	LABEL	BYTE
		db	100h dup(?)	; FOR FLASH ACCESS

_rx_buf         db      RXBUFSIZE dup (?)       ; recv buffer
_recvwp		dw	?		; recv write pointer
_recvrp		dw	?		; recv read pointer

;; recv flowcontrol
_rcvfl		db	?		;

SYSRAM	ENDS

; -----------------------------------------------------------------------------
;  Bootcode for Hardware Reset
;  This routine is called by CPU Hardware Reset
; -----------------------------------------------------------------------------
SETUP	SEGMENT	PARA PUBLIC 'ASMCODE'
	ASSUME CS:SETUP,DS:SYSRAM,ES:SYSRAM,SS:SYSRAM

	org	0

	CLI					; CLI WHILE SETUPING SEQUENCE
	MOV	AX,DGROUP
	MOV	ss,AX
	MOV	sp,offset _ENDSTACK-2

	CALL	InitHardwares			; reset sysrem io resisters
	CALL	InitVector			; reset system vector table
	CALL	InitBios			; reset bios code

	STI					; !!

	jmp	C_INIT_				; jump to LSIC Code

;
; RESET ALL INSIDE SYSTEM IO-REGISTERS
;
InitHardwares	proc	near

	PUSH	AX
	PUSH	DX

	MOV	DX,SCTL
	MOV	AL,SCTL_INITPARAM
	OUT	DX,AL

	;; MUST! BEFORE SETTING OPSEL REG.
	MOV	DX,OPHA
	MOV	AL,OPHA_INITPARAM
	OUT	DX,AL
	MOV	DX,DULA
	MOV	AL,DULA_INITPARAM
	OUT	DX,AL
	MOV	DX,IULA
	MOV	AL,IULA_INITPARAM
	OUT	DX,AL
	MOV	DX,TULA
	MOV	AL,TULA_INITPARAM
	OUT	DX,AL
	MOV	DX,SULA
	MOV	AL,SULA_INITPARAM
	OUT	DX,AL
	MOV	DX,BADR
	MOV	AL,BANK_INITPARAM
	OUT	DX,AL

	MOV	DX,OPSEL
	MOV	AL,OPSEL_INITPARAM
	OUT	DX,AL

	MOV	DX,TCKS
	MOV	AL,TCKS_INITPARAM
	OUT	DX,AL
	MOV	DX,BRC
	MOV	AL,BRC_INITPARAM
	OUT	DX,AL
	
	MOV	DX,WMB0
	MOV	AL,WMB0_INITPARAM
	OUT	DX,AL
	MOV	DX,WCY0
	MOV	AL,WCY0_INITPARAM
	OUT	DX,AL
	MOV	DX,WCY1
	MOV	AL,WCY1_INITPARAM
	OUT	DX,AL

	MOV	DX,WMB1
	MOV	AL,WMB1_INITPARAM
	OUT	DX,AL
	MOV	DX,WAC
	MOV	AL,WAC_INITPARAM
	OUT	DX,AL
	MOV	DX,WCY2
	MOV	AL,WCY2_INITPARAM
	OUT	DX,AL
	MOV	DX,WCY3
	MOV	AL,WCY3_INITPARAM
	OUT	DX,AL
	MOV	DX,WCY4
	MOV	AL,WCY4_INITPARAM
	OUT	DX,AL
;;
;; INITIALIZE PAGE REGISTERS
;;
	MOV	CX,40H
	MOV	DX,PGR_BASE
	MOV	AX,0
@@PAGEREG_SETLOOP:
	OUT	DX,AX
	INC	AX
	ADD	DX,2
	LOOP	@@PAGEREG_SETLOOP
;;
;; INITIALIZE ICU
;; INT0 is Serial Interruput (Edge Trigger)
;; INT1-7 not used(in booting system)
;; AFTER allocated following signal :
;;  INT1 : DMA /TC
;;  INT2 : CD-ROM INTRQ
;;  INT3 : TIMER(TCU2)
;;
	MOV	DX,ICU_BASE
	MOV	AL,00010011B	; LEV=0(EDGE),SINGLE MODE,WRITE IIW4
	OUT	DX,AL		; IIW1
	INC	DX
	MOV	AL,80H		; VECTOR 80H-87H
	OUT	DX,AL		; IIW2
	MOV	AL,00000001B	; 
	OUT	DX,AL		; IIW4
;;
	MOV	AL,11111111B	; ALL Interruput Disable!
	OUT	DX,AL		; IMKW(OCW1)
;;
	PUSH	ES
	PUSH	CX
	PUSH	DI
	MOV	AX,0
	MOV	ES,AX
	MOV	DI,80H*4
	MOV	CX,7
	CLD
@@SetDummyHInt:
	MOV	AX,OFFSET DummyHInt
	STOSW
	MOV	AX,SEG DummyHInt
	STOSW
	LOOP	@@SetDummyHInt
	MOV	AX,OFFSET DummyInt	;; IRQ7(Illegal int handler)
	STOSW
	MOV AX,SEG DummyInt
	STOSW
	
	POP	DI
	POP	CX
	POP	ES

	POP	DX
	POP	AX
	
	RET

InitHardwares	endp

;
; COPY USE VECTORS
; AND INITIALIZE UNUSED VECTOR
;
InitVector	Proc	near

		PUSH	DS
		PUSH	ES
		
		;; copy vectors
		MOV	AX,CS
		MOV	DS,AX
		MOV	AX,0
		MOV	ES,AX
		MOV	SI,OFFSET IntVectTbl
		MOV	DI,0
		MOV	CX,40H*4/2
		CLD
		REPE	MOVSW

		MOV	CX,100H-40H
VECTCOPYLOOP:
		MOV	AX,OFFSET DummyInt
		STOSW
		MOV	AX,CS
		STOSW
		LOOP	VECTCOPYLOOP

		POP	ES
		POP	DS
		
		RET

InitVector	endp

;
; Dummy Software Int CALL
;
DummyInt	proc	far

		iret
		
DummyInt	endp

;
; Dummy Hardware Int CALL
;
DummyHInt	proc	far

		PUSH	DX
		PUSH	AX

		MOV	DX,ICU_BASE
		MOV	AL,20H
		OUT	DX,AL		; EOI

		POP	AX
		POP	DX

		iret

DummyHInt	endp

DivZeroInt	proc	far

		PUSHA
		PUSH	DS
		PUSH	ES

		MOV	DX,ICU_BASE+1	; IMKW
		IN	AL,DX
		AND	AL,1		; Enable INT0?
		JNZ	@@SENDFIN

		MOV	SI,OFFSET DIVZEROMSG
		MOV	AX,CS
		MOV	DS,AX
@@SENDLOOP:
		MOV	DL,[SI]
		CMP	DL,0
		JZ	@@SENDFIN
		MOV	AH,COM_SEND
		INT	COMMBIOS
		OR	AX,AX
		JNZ	@@SENDLOOP
		INC	SI
		JMP	@@SENDLOOP
@@SENDFIN:
		POP	ES
		POP	DS
		POPA
		
		IRET

DIVZEROMSG	DB	'Divided by Zero.',13,10,0

DivZeroInt	endp

IntVectTbl	Label	WORD

		DD	DivZeroInt		; 00
		DD	DummyInt		; 01
		DD	DummyInt		; 02
		DD	DummyInt		; 03
		DD	DummyInt		; 04
		DD	DummyInt		; 05
		DD	DummyInt		; 06
		DD	DummyInt		; 07
		DD	DummyInt		; 08
		DD	DummyInt		; 09
		DD	DummyInt		; 0a
		DD	DummyInt		; 0b
		DD	DummyInt		; 0c
		DD	DummyInt		; 0d
		DD	DummyInt		; 0e
		DD	DummyInt		; 0f
		DD	DummyInt		; 10
		DD	DummyInt		; 11
		DD	DummyInt		; 12
		DD	DummyInt		; 13
		DD	DummyInt		; 14
		DD	DummyInt		; 15
		DD	DummyInt		; 16
		DD	DummyInt		; 17
		DD	DummyInt		; 18
		DD	DummyInt		; 19
		DD	DummyInt		; 1a
		DD	DummyInt		; 1b
		DD	DummyInt		; 1c
		DD	DummyInt		; 1d
		DD	DummyInt		; 1e
		DD	DummyInt		; 1f
		DD	DummyInt		; 20
		DD	DosEmuInt		; 21  Dos Emu
		DD	DummyInt		; 22
		DD	DummyInt		; 23
		DD	DummyInt		; 24
		DD	DummyInt		; 25
		DD	DummyInt		; 26
		DD	DummyInt		; 27
		DD	DummyInt		; 28
		DD	SendChar		; 29  Dos 1Char disp
		DD	DummyInt		; 2a
		DD	DummyInt		; 2b
		DD	DummyInt		; 2c
		DD	DummyInt		; 2d
		DD	DummyInt		; 2e
		DD	DummyInt		; 2f
		DD	CommBiosInt		; 30
		DD	UnicodeInt		; 31
		DD	DrawBiosInt		; 32
		DD	GrphDataInt		; 33
		DD	FlashDataInt		; 34
		DD	ExtAdrInt		; 35
		DD	DummyInt		; 36
		DD	DummyInt		; 37
		DD	DummyInt		; 38
		DD	DummyInt		; 39
		DD	DummyInt		; 3a
		DD	DummyInt		; 3b
		DD	DummyInt		; 3c
		DD	DummyInt		; 3d
		DD	DummyInt		; 3e
		DD	DummyInt		; 3f

InitBios	proc	near

		RET

InitBios	endp

SETUP	ENDS

BIOS	SEGMENT	PARA PUBLIC 'ASMCODE'
	ASSUME CS:BIOS,DS:SYSRAM,ES:SYSRAM,SS:SYSRAM


DosEmuInt	PROC	FAR

		IRET

DosEmuInt	ENDP

;
; INT 29H(DOS : DISP CONSOLE)
;
SendChar	PROC	FAR

		PUSHA

		MOV	DL,AL
@@SENDLOOP:
		MOV	AH,COM_SEND
		INT	COMMBIOS
		OR	AX,AX
		JNZ	@@SENDLOOP

		POPA

		IRET

SendChar	ENDP


CommBiosInt	proc	FAR

		STI	;;(!!)
		PUSH	DS
		PUSH	BX
		MOV	BX,SYSRAM
		MOV	DS,BX
		CMP	AH,COM_GETSTAT
		JA	@@funcret_srl
		XOR	BX,BX
		MOV	BL,AH
		SHL	BX,1
		CALL	WORD PTR CS:FuncTable_srl[BX]
@@funcret_srl:
		POP	BX
		POP	DS

		iret

CommBiosInt	endp

FuncTable_srl	LABEL	WORD
		DW	OFFSET InitRs232c
		DW	OFFSET CloseRs232c
		DW	OFFSET xmitChar
		DW	OFFSET RecvChar
		DW	OFFSET clrrecvbuf
		DW	OFFSET getrecvbufchars
		DW	OFFSET getstat

;
; OPEN RS232C PORT
; SET PARAMETER,INT MASK,INT VECTOR ...
;
InitRs232c	proc	near

		PUSH	DX
		PUSH	DI
		PUSH	ES

		PUSHF
		CLI

		; dat 8bit 1 stopbit no-parity BF=16
		MOV	DX,SIU_BASE+rsSMD
		MOV	AL,01001110B	;
		OUT	DX,AL

		MOV	DX,SIU_BASE+rsSIMK
		MOV	AL,00000010B	; Enable RXINT,Disable TXINT
		OUT	DX,AL

		CLD
		MOV	AX,0
		MOV	ES,AX
		MOV	DI,80H*4
		MOV	AX,OFFSET __proc_intr
		STOSW
		MOV	AX,CS
		STOSW

		MOV	DX,SIU_BASE+rsSCM
		MOV	AL,00110101B	; RTS ON,RX/TX Enable,Break off,DTR ON
		OUT	DX,AL

		MOV	AL,_ON
		MOV	_rcvfl,AL	; Recv OK

		XOR	AX,AX
		MOV	_recvwp,AX	; Init Recv Point
		MOV	_recvrp,AX

		;; ICU Mask Off
		MOV	DX,ICU_BASE+1	; IMKW
		IN	AL,DX
		AND	AL,11111110B	; Enable INT0
		OUT	DX,AL

		POPF

		POP	ES
		POP	DI
		POP	DX

		RET

InitRs232c	endp

CloseRs232c	proc	near

		PUSHF
		CLI

		MOV	AL,_ON
		MOV	_rcvfl,AL	; Recv OK

		XOR	AX,AX
		MOV	_recvwp,AX	; Init Recv Point
		MOV	_recvrp,AX

		MOV	DX,SIU_BASE+rsSIMK
		MOV	AL,00000011B	; Disable RXINT/TXINT
		OUT	DX,AL

		;; ICU Mask Off
		MOV	DX,ICU_BASE+1	; IMKW
		IN	AL,DX
		OR	AL,1		; Disable INT0
		OUT	DX,AL

		POPF
		
		RET

CloseRs232c	endp


; -------------------------------------------------------------------------
; RECV 1 DATA
; *inregs
; NONE
; *outregs
; DX : RECEIVED DATA
; AX : 0 : OK , 1 : NG
; -------------------------------------------------------------------------
recvchar	proc	near

		PUSHF
		CLI
		MOV	AX,_recvwp
		MOV	BX,_recvrp
		POPF
		SUB	AX,BX
		JZ	@@recvbyte_nodata
		AND	AX,RXBUFSIZE-1
		MOV	DX,AX
		MOV	AL,_rx_buf[BX]
		INC	BX
		AND	BX,RXBUFSIZE-1
		MOV	_recvrp,BX
		CMP	DX,FLOW_ON_POINT
		JNZ	@@recvbyte_gotdata
		CMP	_rcvfl,_ON
		JZ	@@recvbyte_gotdata
		CALL	recv_rscheck
@@recvbyte_gotdata:
		MOV	AH,0		; CLEAR HIGH BYTE
		MOV	DX,AX		; SET DATA
		XOR	AX,AX		; OK

		RET
;
; NOT EXIST RECEIVED DATA
@@recvbyte_nodata:
		XOR	DX,DX
		MOV	AX,-1
		
		RET

recvchar	endp

;-----------------------------------------------------------------
; RS 0N AND FLAGS ON
;-----------------------------------------------------------------
recv_rscheck	proc	near

		PUSH	AX
		MOV	_rcvfl,_ON
;; rts on
		PUSH	DX
		MOV	DX,SIU_BASE+rsSCM
		IN	AL,DX
		OR	AL,00100000B		; RTS ON
		OUT	DX,AL
		POP	DX

		POP	AX

		RET

recv_rscheck	endp

; -------------------------------------------------------------------------
; send 1byte
; *inregs
; DL : send data
; *outregs
; AX : result  0 : ok , else ng
; -------------------------------------------------------------------------
xmitchar	proc	near

		PUSH	DX
		MOV	AH,DL			; AH=data
		MOV	DX,SIU_BASE+rsSST
		IN	AL,DX
		AND	AL,1			; 
		JZ	@@xmitbyte_err
		MOV	DX,SIU_BASE+rsSTB
		MOV	AL,AH
		OUT	DX,AL			; write Buffer
		XOR	AX,AX			; ok!
		POP	DX

		RET
@@xmitbyte_err:
		MOV	AX,-1			; send err
		POP	DX

		RET

xmitchar	endp

; ------------------------------------------------
; GET STATUS
; *inregs
;  NONE
; *outregs
;  AL : STATUS
; ------------------------------------------------
getstat		proc	near

		PUSH	DX
		MOV	DX,SIU_BASE+rsSST
		IN	AL,DX
		MOV	AH,0
		POP	DX

		RET

getstat		endp


;------------------------------------------------
; Get number of datas IN receive buffer
; *inregs
;    none
; *outregs.
;    AX   number of datas
;------------------------------------------------
getrecvbufchars	proc	near

		PUSHF
		CLI
		MOV	AX,_recvwp	;len:=rcvwp-rcvrp
		SUB	AX,_recvrp
		POPF
		AND	AX,RXBUFSIZE-1
		RET

getrecvbufchars	endp

;------------------------------------------------------------
; Clear receive buffer
; *inregs.
;    none
; *outregs.
;    none
;------------------------------------------------------------
clrrecvbuf	proc	near

		PUSHF
		CLI
		PUSH	AX
		XOR	AX,AX
		MOV	_recvrp,AX
		MOV	_recvwp,AX
		CMP	_rcvfl,_ON
		MOV	_rcvfl,_ON
                JZ      @@initrecvbuf_setintmask
		CALL	recv_rscheck
@@initrecvbuf_setintmask:
		POP	AX
		POPF

		RET

clrrecvbuf	endp

; -----------------------------------------------------------------
;  rs-232c interrupt routine for V53A inside siu
; -----------------------------------------------------------------
__proc_intr	proc    far

		PUSH	AX
		PUSH	BX
		PUSH	DX
		PUSH    DS

		MOV	AX,SYSRAM
		MOV	DS,AX

		MOV	AH,_OFF			;for speed!

		MOV	DX,SIU_BASE+rsSST
		IN	AL,DX
		AND	AL,00000010b		;exist new received data ?
		JZ	@@Finish_Interrupt

		MOV	DX,SIU_BASE+rsSRB	;
		IN	AL,DX			;
		MOV	BX,_recvwp		;
		MOV	_rx_buf[BX],AL		;
		INC	BX			;pointer++
		AND	BX,RXBUFSIZE-1		;
		CMP	BX,_recvrp		;check overflow
		JZ	@@Finish_Interrupt	;if overflow then not write data
		MOV	_recvwp,BX		;refresh pointer
		CMP	_rcvfl,AH
		je	@@Finish_Interrupt	;
		SUB	BX,_recvrp		;
		AND	BX,RXBUFSIZE-1		;
		CMP	BX,FLOW_OFF_POINT	;chk flow
		jne	@@Finish_Interrupt	;
		MOV	_rcvfl,AH		;recv flag off
		MOV	DX,SIU_BASE+rsSCM
		IN	AL,DX
		AND	AL,11011111B
		OUT	DX,AL			; RTS OFF

@@Finish_Interrupt:
		MOV	DX,ICU_BASE		; MASTER EOI
		MOV	AL,20h
		OUT	DX,AL

		POP     DS
		POP	DX
		POP	BX
		POP	AX

		iret

__proc_intr	endp

;
; Flash memory data save/load module.
; SAVE BLOCK = 254 bytes
; 
FlashDataInt	proc	far

		STI	;;(!!)
		PUSH	DS
		PUSH	BX
		MOV	BX,SYSRAM
		MOV	DS,BX
		CMP	AH,FLASH_SAVEDATA
		JA	@@funcret_flash
		XOR	BX,BX
		MOV	BL,AH
		SHL	BX,1
		CALL	WORD PTR CS:FuncTable_flash[BX]
@@funcret_flash:
		POP	BX
		POP	DS

		iret

FlashDataInt	endp

FuncTable_flash	LABEL	WORD
		DW	_LOAD_DATA
		DW	_SAVE_DATA

; ---------------------------------------------------------------

;
; ES:DI <- STORE ADDRESS
;
_LOAD_DATA	PROC	NEAR

		PUSH	SI
		PUSH	DI
		PUSH	DS

		CLI				;; !!
		MOV	AX,7000H		;;
		CALL	_SET_PAGEAREA
		CALL	_CHANGE_PAGE		;; E0000-EFFFF <- FLASH HIDE-BANK
		CALL	_SEARCH_LASTBLOCK	;;
		JC	@@LOADDATA_NODATA

		MOV	SI,AX
		MOV	AX,FLASH_PAGINGSEG
		MOV	DS,AX
		MOV	CX,254 / 2
		CLD
		REPE	MOVSW
		XOR	AX,AX
		JMP	@@LOADDATA_EXIT

@@LOADDATA_NODATA:
		MOV	AX,-1
@@LOADDATA_EXIT:
		CALL	_RESTORE_PAGE		;; E0000-EFFFF RESTORE.

		POP	DS
		POP	DI
		POP	SI

		RET

_LOAD_DATA	ENDP

;
; ES:DI -> SAVE DATA ADDRESS
;
_SAVE_DATA	PROC	NEAR

		PUSH	CX
		PUSH	SI
		PUSH	DI
		PUSH	DS
		PUSH	ES

		PUSH	ES
		PUSH	DI

		CLI				;; !!
		MOV	AX,7000H		;;
		CALL	_SET_PAGEAREA
		CALL	_CHANGE_PAGE		;; E0000-EFFFF <- FLASH HIDE-BANK
		CALL	_COPY_PROC		;; COPY FLASH WRITE CODE TO STATIC-RAM
		CALL	_SEARCH_LASTBLOCK	;; RESULT IS CARRY-F

		MOV	DI,AX			;; SET WRITE ADDRESS (PREV BLOCK)
		JC	@@SAVEDATA_ALLBLANK

		CMP	AX,0FF00H		;; LAST BLOCK?
		JNZ	@@SAVEDATA_NORMAL
;;
		MOV	AX,FLASH_PAGINGSEG
		MOV	DX,0
		CALL	FAR PTR _EXECUTION_AREA	;; EXECUTE ERASE SECTOR
		CMP	AX,0
		JNZ	@@SAVEDATA_ERR

@@SAVEDATA_NORMAL:
		ADD	DI,100H			;; WRITE BLOCK.
@@SAVEDATA_ALLBLANK:
		POP	SI
		POP	DS

		MOV	AX,FLASH_PAGINGSEG
		MOV	ES,AX
		MOV	CX,254 / 2
		CLD
@@SAVEDATA_LOOP:
		LODSW
		CMP	AX,0FFFFH
		JZ	@@SAVEDATA_SKIP
		CALL	FAR PTR _EXECUTION_AREA+2	; WRITE DATA
		CMP	AX,0
		JNZ	@@SAVEDATA_ERR1
@@SAVEDATA_SKIP:
		ADD	DI,2
		LOOP	@@SAVEDATA_LOOP

		MOV	AX,0				; NULL CODE
		CALL	FAR PTR _EXECUTION_AREA+2	; WRITE DATA
		CMP	AX,0
		JNZ	@@SAVEDATA_ERR1
		XOR	AX,AX
@@SAVEDATA_EXIT:
		CALL	_RESTORE_PAGE		;; E0000-EFFFF RESTORE.

		POP	ES
		POP	DS
		POP	DI
		POP	SI
		POP	CX

		RET

@@SAVEDATA_ERR:
		POP	SI
		POP	DS
@@SAVEDATA_ERR1:
		MOV	AX,-1
		JMP	@@SAVEDATA_EXIT

_SAVE_DATA	ENDP


_COPY_PROC	PROC	NEAR

		PUSH	AX
		PUSH	CX
		PUSH	SI
		PUSH	DI
		PUSH	DS
		PUSH	ES

		MOV	AX,CS
		MOV	DS,AX
		MOV	AX,SYSRAM
		MOV	ES,AX
		MOV	SI,OFFSET __FLASH_RAMEXECUTION_CODE_TOP
		MOV	DI,OFFSET _EXECUTION_AREA
		MOV	CX,(COPYSIZE + 1) / 2
		CLD
		REPE	MOVSW
		
		POP	ES
		POP	DS
		POP	DI
		POP	SI
		POP	CX
		POP	AX
		
		RET

_COPY_PROC	ENDP

COPYSIZE = __FLASH_RAMEXECUTION_CODE_BOTTOM-OFFSET __FLASH_RAMEXECUTION_CODE_TOP+1

_SEARCH_LASTBLOCK	PROC	NEAR

		PUSH	ES
		PUSH	DS
		PUSH	DI
		PUSH	SI
		PUSH	CX

		MOV	AX,FLASH_PAGINGSEG
		MOV	ES,AX
		MOV	DI,0FFFEH
		MOV	CX,8000H
		MOV	AX,0FFFFH		;; NULL CHECK
		PUSHF
		STD
		REPE	SCASW
		POPF
		MOV	AX,DI
		CMP	CX,0
		JZ	@@SEARCHLASTBLOCK_NOBLOCK

		AND	AX,0FF00H
		CLC
		JMP	@@SEARCHLASTBLOCK_EXIT

@@SEARCHLASTBLOCK_NOBLOCK:
		STC
		MOV	AX,0
@@SEARCHLASTBLOCK_EXIT:

		POP	CX
		POP	SI
		POP	DI
		POP	DS
		POP	ES

		RET

_SEARCH_LASTBLOCK	ENDP

;
; FLASH-MEMORY PAGING AREA IS $E000:0000 - $E000:FFFF
; 64KBYTE@1PAGE
;
; AX = UPPER SEGMENT (0000H,1000H,...)
;
_SET_PAGEAREA	PROC	NEAR

		PUSH	CX
		PUSH	DX

		AND	AX,07000H	; 3BIT ONLY (FLASH WRITE AREA SIZE = 80000H)
		SHR	AX,10		; 
		ADD	AX,40H		; UPPER 100000H
		MOV	DX,PGR_BASE+(FLASH_PAGINGSEG SHR 12)*4*2

		MOV	CX,4
@@SETPAGEAREA_LOOP:
		OUT	DX,AX
		ADD	DX,2
		INC	AX
		LOOP	@@SETPAGEAREA_LOOP

		POP	DX
		POP	CX

		RET

_SET_PAGEAREA	ENDP


;; ---------------------------------------------------------------------------
_CHANGE_PAGE	PROC	FAR		;!!

		BRKXA	EXTADRVECT

;;;		RET		BRKXA Inst. execute FAR JUMP AT VECT(EXTVECT),
;;;                             AND EXTADR-ENABLE-FLAG(XA FLAG) SET 1.

_CHANGE_PAGE	ENDP

_RESTORE_PAGE	PROC	FAR

		RETXA	EXTADRVECT

;;;		RET		RETXA Inst. execute FAR JUMP AT VECT(EXTVECT),
;;;                             AND EXTADR-ENABLE-FLAG(XA FLAG) SET 0.

_RESTORE_PAGE	ENDP

;
; BRKXA,RETXA JUMP ADDRESS
;
EXTADRINT	PROC	FAR

		RET		;; FAR RETURN AT THIS POINT.

EXTADRINT	ENDP
;; ---------------------------------------------------------------------------

; ---------------------------------------------------------------------
;  FLASH ACCESS CODE(ERASE/WRITE)
;  THIS ROUTINES ARE EXPANDED RAM AREA WHEN FLASH-MEMORY ACCESSED.
; ---------------------------------------------------------------------

__FLASH_RAMEXECUTION_CODE_TOP	label	byte

		NOJUMPS		;; 
;
; Jmp call table
;
		JMP	SHORT ERASE_SECTOR
		JMP	SHORT WRITE_WORD

;
; Program Flash WORD Write
; inparam  AX    = Write DATA
;          ES:DI = address of writing position
; outparam AX    = RESULT
;          = 0  SUCCESS
;          = -1 FAIL

WRITE_WORD	PROC	FAR

		PUSHF
		CLI
		MOV	WORD PTR ES:[0AAAAh],0F0F0h	; 
		MOV	WORD PTR ES:[0AAAAh],0AAAAh
		MOV	WORD PTR ES:[05554h],05555h
		MOV	WORD PTR ES:[0AAAAh],0A0A0h
		MOV	WORD PTR ES:[DI],AX
;
		CALL	WAIT_TOGGLE_BIT
		MOV	WORD PTR ES:[0AAAAh],0F0F0h	; 
		POPF

		RET

WRITE_WORD	ENDP

; -----------------------------------------------------------------
; ERASE SELECTED SECTOR
; inparam ax = erase segment address
;         dx = erase sector address
; outparam ax = result
;             =0  success
;             =-1 fail
ERASE_SECTOR	PROC	FAR
;
		PUSH	ES
		PUSH	DI

		MOV	ES,AX
		MOV	DI,DX
;;
		PUSHF
		CLI
		MOV	WORD PTR ES:[0AAAAh],0F0F0h	; RESET COMMAND
;
		MOV	WORD PTR ES:[0AAAAh],0AAAAh
		MOV	WORD PTR ES:[05554h],05555h
		MOV	WORD PTR ES:[0AAAAh],08080h
		MOV	WORD PTR ES:[0AAAAh],0AAAAh
		MOV	WORD PTR ES:[05554h],05555h
		MOV	WORD PTR ES:[DI],03030h		; ERASE SECTOR COMMAND
;
		CALL	WAIT_TOGGLE_BIT
		MOV	WORD PTR ES:[0AAAAh],0F0F0h	; RESET COMMAND
		POPF
;
		POP	DI
		POP	ES
;
		RET

ERASE_SECTOR	ENDP

; -----------------------------------------------------------------
; CHECK TOGGLE BIT STATE
;
; inparam   ES:DI  = address of toggle process
; outparam  AX     = result (0:success, -1:process err)
;
WAIT_TOGGLE_BIT	PROC	NEAR

		PUSH	DX

@@WAIT_BUSY_LOOP:
		MOV	AX,ES:[DI]
		MOV	DX,ES:[DI]

		PUSH	AX		; !!
		AND	AX,01000000B	; FIRST READ
		AND	DX,01000000B	; SECOND READ
		XOR	AX,DX		; DQ6 IS SAME?
		JZ	@@WAIT_NOBUSY_OK0
		POP	AX		; !!
		TEST	AX,00100000B	; DQ5(CHECK TIMEOVER)
		JZ	@@WAIT_BUSY_LOOP
;
		MOV	AX,ES:[DI]
		MOV	DX,ES:[DI]
		AND	AX,01000000B	; FIRST  READ
		AND	DX,01000000B	; SECOND READ
		XOR	AX,DX		; DQ6 IS SAME?
		JZ	@@WAIT_NOBUSY_OK1

		MOV	AX,-1
		JMP	@@WAIT_NOBUSY_EXIT

@@WAIT_NOBUSY_OK0:
		POP	AX
@@WAIT_NOBUSY_OK1:
		XOR	AX,AX
@@WAIT_NOBUSY_EXIT:
;
		POP	DX
		RET

WAIT_TOGGLE_BIT	ENDP

__FLASH_RAMEXECUTION_CODE_BOTTOM	label	byte

BIOS	ENDS

;
; Hardware Reset point FFFF:0000
;
BOOT	SEGMENT	BYTE	PUBLIC 'BOOTCODE'
	ASSUME CS:BOOT

	org	0

	DB	0EAh		;; JMP FAR 8000:0000
	DW	0
	DW	SETUP

BOOT	ENDS

END
