/* MPROM Firmware Code version 1.0 (LCD Drawing library)             */
/* programmed by M.yanagisawa(MoroboshiLUM)  elfin@mth.biglobe.ne.jp */
/* Copyright (c) 2000 by M.yanagisawa. All rigths reserved.          */

#ifdef __BORLANDC__

#include <dos.h>

#else

#include <stdrom.h>

#endif

#include "drawlib.h"

void Grph_Init(void)
{
   union REGS i,o;

#ifdef __BORLANDC__
  InitSimpleDraw();
#else
  i.h.ah=DRAW_INIT;
  int86(DRAW_BIOS,&i,&o);
#endif
}

void Grph_SetArea(int x1,int y1,int x2,int y2)
{
   TSETCLIPAREA_PACS pac;
   union REGS i,o;
   struct SREGS sr;

  pac.sx=x1; pac.sy=y1;
  pac.ex=x2; pac.ey=y2;
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   SetClipArea();
#else
   i.h.ah=DRAW_SETCLIPAREA;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_SetDepth(int depth)
{
   union REGS i,o;

#ifdef __BORLANDC__
  asm {
    mov al,byte ptr depth
  }
  SelectPattern();
#else
  i.h.ah=DRAW_SETPATTERN;
  i.h.al=depth;
  int86(DRAW_BIOS,&i,&o);
#endif
}

void Grph_Pset(int x,int y,int mode,int size)
{
   union REGS i,o;
   
#ifdef __BORLANDC__
   asm {
     mov bx,x
     mov dx,y
     mov ch,byte ptr size
     mov cl,byte ptr mode
   };
   PutPixel();
#else
   i.h.ah=DRAW_POINT;
   i.x.bx=x;   i.x.dx=y;
   i.h.ch=size; i.h.cl=mode;
   int86(DRAW_BIOS,&i,&o);
#endif
}

void Grph_Line(int x1,int y1,int x2,int y2,int mode,int size,int style)
{
   TDRAWLINE_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.x1=x1;  pac.y1=y1;
   pac.x2=x2;  pac.y2=y2;
   pac.mode=mode; pac.size=size; pac.style=style;
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   DrawLine();
#else
   i.h.ah=DRAW_LINE;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_Box(int x1,int y1,int x2,int y2,int mode,int size,int style)
{
   TDRAW_COMMON_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.x1=x1;  pac.y1=y1;
   pac.x2=x2;  pac.y2=y2;
   pac.mode=mode; pac.size=size; pac.style=style;
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   DrawBox();
#else
   i.h.ah=DRAW_BOX;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_FillArea(int x1,int y1,int x2,int y2,int mode,int depth)
{
   TFILLAREA_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.sx=x1;  pac.sy=y1;
   pac.ex=x2;  pac.ey=y2;
   
#ifdef __BORLANDC__
   asm {
     mov ax,depth
     shl al,4
     mov bx,mode
     and bl,0fh
     or  al,bl
     lea bx,pac
     push ss
     pop es
   };
   FillArea();
#else
   i.h.ah=DRAW_FILLAREA;
   i.h.al=(depth << 4) | (mode & 0xf);
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_GetImage(int x,int y,int width,int height,void far *body)
{
   TGETIMAGE_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.x=x;  pac.y=y;
   pac.width=width;  pac.height=height;
   pac.ofs=FP_OFF(body); pac.seg=FP_SEG(body);
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   DrawGetImage();
#else
   i.h.ah=DRAW_GETIMAGE;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_PutImage(int x,int y,int width,int height,int mode,void far *body)
{
   TPUTIMAGE_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.x=x;  pac.y=y;
   pac.width=width;  pac.height=height; pac.mode=mode;
   pac.ofs=FP_OFF(body); pac.seg=FP_SEG(body);
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   DrawPutImage();
#else
   i.h.ah=DRAW_PUTIMAGE;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

void Grph_PutKanji(int x,int y,unsigned int code,int mode)
{
   TPUTKANJI_PACS pac;
   union REGS i,o;
   struct SREGS sr;
   
   pac.x=x;  pac.y=y;
   pac.code=code;  pac.mode=mode;
#ifdef __BORLANDC__
   asm {
     lea bx,pac
     push ss
     pop es
   };
   DrawPutKanji();
#else
   i.h.ah=DRAW_PUTKANJI;
   i.x.bx=FP_OFF((void far *)&pac);
   sr.es=FP_SEG((void far *)&pac);
   int86x(DRAW_BIOS,&i,&o,&sr);
#endif
}

#ifdef __BORLANDC__

BYTE zerotbl[]={
	0x00, 0x0F, 0xF0, 0x00,
	0x00, 0x7F, 0xFE, 0x00,
	0x00, 0xFF, 0xFF, 0x00,
	0x01, 0xFF, 0xFF, 0x80,
	0x03, 0xFF, 0xFF, 0xC0,
	0x07, 0xFF, 0xFF, 0xE0,
	0x0F, 0xFF, 0xFF, 0xF0,
	0x0F, 0xFF, 0xFF, 0xF0,
	0x1F, 0xFC, 0x3F, 0xF8,
	0x1F, 0xF0, 0x0F, 0xF8,
	0x1F, 0xF0, 0x0F, 0xF8,
	0x3F, 0xE0, 0x07, 0xFC,
	0x3F, 0xE0, 0x07, 0xFC,
	0x3F, 0xE0, 0x07, 0xFC,
	0x3F, 0xE0, 0x03, 0xFC,
	0x3F, 0xC0, 0x03, 0xFC,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFE,
	0x7F, 0xC0, 0x03, 0xFC,
	0x3F, 0xC0, 0x07, 0xFC,
	0x3F, 0xE0, 0x07, 0xFC,
	0x3F, 0xE0, 0x07, 0xFC,
	0x3F, 0xE0, 0x07, 0xFC,
	0x1F, 0xF0, 0x0F, 0xF8,
	0x1F, 0xF0, 0x0F, 0xF8,
	0x1F, 0xFC, 0x3F, 0xF8,
	0x0F, 0xFF, 0xFF, 0xF0,
	0x0F, 0xFF, 0xFF, 0xF0,
	0x07, 0xFF, 0xFF, 0xE0,
	0x03, 0xFF, 0xFF, 0xC0,
	0x01, 0xFF, 0xFF, 0x80,
	0x00, 0xFF, 0xFF, 0x00,
	0x00, 0x7F, 0xFE, 0x00,
	0x00, 0x0F, 0xF0, 0x00,
	0x00, 0x00, 0x00, 0x00
};

BYTE colontbl[] = {
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x1F, 0xF0,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00
};

BYTE minustbl[] = {
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0xFF, 0xFF,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00,
	0x00, 0x00
};
#endif

BOOL Grph_GetImageData(int num,int *x,int *y,void far **data)
{
   union REGS i,o;

#ifdef __BORLANDC__
/* test code */
   if (num>9) {
     *x=16;
     *y=48;
     if (num==10) *data=(void far *)colontbl;
     else *data=(void far *)minustbl;
   } else {
     *x=32;
     *y=48;
     *data=(void far *)zerotbl;
   }
   return(TRUE);
#else
   i.x.ax=num;
   int86(GRPHDATA_BIOS,&i,&o);
   if (!o.x.cx) return(FALSE);
   *x=o.x.cx; *y=o.x.bx;
   *data=(void far *)MK_FP(o.x.dx,o.x.ax);
   return(TRUE);
#endif
}
